from euphorie.deployment.setuphandlers import setupSecureSessionCookie
from ftw.upgrade import UpgradeStep
from plone import api


class ActivateSecureSessionCookie(UpgradeStep):
    """Activate secure session cookie."""

    def __call__(self):
        site = api.portal.get()
        setupSecureSessionCookie(site)



from euphorie.client import utils
from euphorie.client.browser.webhelpers import WebHelpers
from euphorie.client.country import ClientCountry
from euphorie.client.interfaces import IClientSkinLayer
from euphorie.client.sector import ClientSector
from euphorie.client.tests.utils import testRequest
from euphorie.client.utils import locals
from euphorie.content.survey import Survey
from euphorie.testing import EuphorieIntegrationTestCase
from plone import api
from zope.interface import alsoProvides

import unittest


class MockRequest:
    form = {}

    def __init__(self, agent=None):
        self.__headers = {}
        if agent is not None:
            self.__headers["User-Agent"] = agent

    def purge_memoize(self):
        """Clean up the memoize cache."""
        try:
            self.__annotations__.pop("plone.memoize", None)
        except AttributeError:
            pass

    def get_header(self, key, default):
        return self.__headers.get(key, default)


class MockSession:
    def __init__(self, account=None):
        self.account = account


class TestURLs(EuphorieIntegrationTestCase):
    def setUp(self):
        super().setUp()
        # Set locals
        request = testRequest()
        locals.request = request
        self.loginAsPortalOwner()
        self.client = self.portal.client
        # Add survey
        self.client._setOb("en", ClientCountry("en"))
        country = self.client["en"]
        country._setOb("sector", ClientSector("sector"))
        country["sector"].title = "Test sector"
        country["sector"].id = "sector"
        sector = country["sector"]
        survey = Survey("survey")
        survey.title = "Test Survey"
        survey.introduction = "This is a survey that is well suited for tests"
        survey.language = "en"
        sector._setOb("survey", survey)

    def _get_view(self, context):
        request = self.request.clone()
        alsoProvides(request, IClientSkinLayer)
        return api.content.get_view("webhelpers", context, request)

    def testBaseURL(self):
        country = self.client["en"]
        survey = self.client["en"]["sector"]["survey"]

        view = self._get_view(self.client)
        self.assertEqual(view._base_url(), self.client.absolute_url())
        view = self._get_view(country)
        self.assertEqual(view._base_url(), country.absolute_url())

        view = self._get_view(survey)
        self.assertEqual(view._base_url(), survey.absolute_url())


class WebhelperTests(EuphorieIntegrationTestCase):
    def _createView(self, agent=None):
        return WebHelpers(self.portal.client, MockRequest(agent))

    def testIsIphone_NoUserAgent(self):
        self.assertEqual(self._createView().is_iphone, False)

    def testIsIphone_IE6(self):
        view = self._createView(
            agent="Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.2; "
            "SV1; .NET CLR 1.1.4322; .NET CLR 2.0.50727; InfoPath.1)"
        )
        self.assertEqual(view.is_iphone, False)

    def testIsIphone_iPhone30(self):
        view = self._createView(
            agent="Mozilla/5.0 (iPod; U; CPU iPhone OS 3_0 like Mac OS X; "
            "en-us) AppleWebKit/528.18 (KHTML, like Gecko) "
            "Version/4.0 Mobile/7A341 Safari/528.16"
        )
        self.assertEqual(view.is_iphone, True)


class HasTextTests(unittest.TestCase):
    def testNone(self):
        self.assertEqual(utils.HasText(None), False)

    def testEmpty(self):
        self.assertEqual(utils.HasText(""), False)

    def testSimpleText(self):
        self.assertEqual(utils.HasText("Hello, World"), True)

    def testWhitespaceOnly(self):
        self.assertEqual(utils.HasText("   &nbsp;"), False)

    def testTagsOnly(self):
        self.assertEqual(utils.HasText("<strong></strong>"), False)

    def testWhitespaceInTag(self):
        self.assertEqual(utils.HasText("<strong>  </strong>"), False)

    def testTagsAndText(self):
        self.assertEqual(utils.HasText("<strong>STRONG</strong>"), True)


class RandomStringTests(unittest.TestCase):
    def testOutputChanges(self):
        self.assertNotEqual(utils.randomString(), utils.randomString())

    def testLength(self):
        self.assertEqual(len(utils.randomString(5)), 5)



from ..risk import evaluation_algorithm
from ..risk import IFrenchEvaluation
from ..risk import IFrenchRisk
from ..risk import IKinneyEvaluation
from ..risk import IKinneyRisk
from ..risk import IRisk
from ..solution import ISolution
from ..utils import DragDropHelper
from ..utils import getTermTitleByValue
from Acquisition import aq_base
from Acquisition import aq_inner
from Acquisition import aq_parent
from Acquisition.interfaces import IAcquirer
from euphorie.content.survey import get_tool_type
from euphorie.content.utils import get_tool_type_default
from euphorie.content.utils import IToolTypesInfo
from euphorie.content.utils import parse_scaled_answers
from plone import api
from plone.dexterity.browser.add import DefaultAddForm
from plone.dexterity.browser.add import DefaultAddView
from plone.dexterity.browser.edit import DefaultEditForm
from plone.dexterity.interfaces import IDexterityFTI
from plone.memoize.instance import memoize
from Products.Five import BrowserView
from Products.statusmessages.interfaces import IStatusMessage
from z3c.form.form import applyChanges
from zope.component import createObject
from zope.component import getUtility
from zope.interface import alsoProvides


class RiskView(BrowserView, DragDropHelper):
    @property
    @memoize
    def my_context(self):
        return aq_inner(self.context)

    @property
    def module_title(self):
        return aq_parent(self.context).title

    @property
    def evaluation_algorithm(self):
        return evaluation_algorithm(self.my_context)

    @property
    def risk_type(self):
        return getTermTitleByValue(IRisk["type"], self.my_context.type)

    @property
    def tool_type_data(self) -> dict:
        """There is an utility that provides the tool types information.

        The information are stored in a dictionary with the tool type as key.

        The default utility return the dict `euphorie.content.utils.TOOL_TYPES`.

        This is not directly used by this package, but custom code will use this.
        """
        tool_types_info = getUtility(IToolTypesInfo)()
        tool_type = get_tool_type(self.my_context)
        try:
            return tool_types_info[tool_type]
        except KeyError:
            return tool_types_info[get_tool_type_default()]

    @property
    def solutions(self):
        return [
            {
                "id": solution.id,
                "url": solution.absolute_url(),
                "description": solution.description,
            }
            for solution in self.my_context.values()
            if ISolution.providedBy(solution)
        ]

    @property
    def evaluation_method(self):
        return getTermTitleByValue(
            IRisk["evaluation_method"], self.my_context.evaluation_method
        )

    @property
    def default_priority(self):
        return getTermTitleByValue(
            IKinneyRisk["default_priority"], self.my_context.default_priority
        )

    @property
    def default_severity(self):
        return getTermTitleByValue(
            IFrenchEvaluation["default_severity"], self.my_context.default_severity
        )

    @property
    def default_frequency(self):
        if self.evaluation_algorithm == "french":
            return getTermTitleByValue(
                IFrenchEvaluation["default_frequency"],
                self.my_context.default_frequency,
            )
        else:
            return getTermTitleByValue(
                IKinneyEvaluation["default_frequency"],
                self.my_context.default_frequency,
            )

    @property
    def default_probability(self):
        return getTermTitleByValue(
            IKinneyEvaluation["default_probability"],
            self.my_context.default_probability,
        )

    @property
    def default_effect(self):
        return getTermTitleByValue(
            IKinneyEvaluation["default_effect"], self.my_context.default_effect
        )

    @property
    @memoize
    def portal_transforms(self):
        return api.portal.get_tool("portal_transforms")

    def get_safe_html(self, text):
        data = self.portal_transforms.convertTo(
            "text/x-html-safe", text, mimetype="text/html"
        )
        return data.getData()

    @property
    @memoize
    def scaled_answers(self):
        """Get values and answers if the scaled_answers field is used.

        This returns a list of dictionaries.
        """
        context = self.my_context
        if not getattr(context, "use_scaled_answer", False):
            return []
        return parse_scaled_answers(context.scaled_answers)


class RiskFieldsetOrderingMixin:

    order = [
        "header_identification",
        "header_evaluation",
        "header_main_image",
        "header_secondary_images",
        "header_additional_content",
    ]

    def updateFields(self):
        """
        Override the parent method to sort fieldsets based on the predefined order.

        The method uses the `order` attribute to determine the position of each
        fieldset. If a fieldset's label is not found in the `order` list, it is
        placed at the end. Sorting is performed in-place on the `self.groups` list.
        """
        super().updateFields()

        def index_of_group(group):
            try:
                return self.order.index(group.label)
            except ValueError:
                return len(self.order)

        self.groups.sort(key=index_of_group)


class AddForm(RiskFieldsetOrderingMixin, DefaultAddForm):
    portal_type = "euphorie.risk"
    default_fieldset_label = None

    @property
    @memoize
    def evaluation_algorithm(self):
        return evaluation_algorithm(aq_inner(self.context))

    @property
    def schema(self):
        if self.evaluation_algorithm == "french":
            return IFrenchRisk
        else:
            return IKinneyRisk

    def updateWidgets(self):
        super().updateWidgets()
        self.widgets["title"].addClass("span-7")
        self.widgets["existing_measures"].mode = "hidden"

    def create(self, data):
        # This is mostly a direct copy of
        # :py:meth:`plone.dexterity.browser.add.DefaultAddForm.create`,
        # extended to apply the right interface.
        fti = getUtility(IDexterityFTI, name=self.portal_type)
        container = aq_inner(self.context)
        content = createObject(fti.factory)
        alsoProvides(content, self.schema)
        if hasattr(content, "_setPortalTypeName"):
            content._setPortalTypeName(fti.getId())
        if IAcquirer.providedBy(content):
            content = content.__of__(container)
        applyChanges(self, content, data)
        for group in self.groups:
            applyChanges(group, content, data)
        return aq_base(content)


class AddView(DefaultAddView):
    form = AddForm


class EditForm(RiskFieldsetOrderingMixin, DefaultEditForm):
    portal_type = "euphorie.risk"
    default_fieldset_label = None

    @property
    @memoize
    def my_context(self):
        return aq_inner(self.context)

    @property
    @memoize
    def evaluation_algorithm(self):
        return self.my_context.evaluation_algorithm()

    @property
    def schema(self):
        if self.evaluation_algorithm == "french":
            return IFrenchRisk
        else:
            return IKinneyRisk

    @property
    def use_existing_measures(self):
        return api.portal.get_registry_record(
            "euphorie.use_existing_measures", default=False
        )

    @property
    @memoize
    def portal_transforms(self):
        return api.portal.get_tool("portal_transforms")

    def get_safe_html(self, text):
        data = self.portal_transforms.convertTo(
            "text/x-html-safe", text, mimetype="text/html"
        )
        return data.getData()

    @property
    def tool_type(self):
        return get_tool_type(self.my_context)

    def updateWidgets(self):
        super().updateWidgets()
        self.widgets["title"].addClass("span-7")
        tt = getUtility(IToolTypesInfo)
        if not (
            self.use_existing_measures and self.tool_type in tt.types_existing_measures
        ):
            self.widgets["existing_measures"].mode = "hidden"
        else:
            self.widgets["existing_measures"].mode = "display"
        for fname in ("description", "legal_reference"):
            value = self.widgets[fname].value or ""
            safe_value = self.get_safe_html(value)
            if value != safe_value:
                self.widgets[fname].value = safe_value

    def extractData(self, setErrors=True):
        data = super().extractData(setErrors)
        if data[0]["evaluation_method"] == "fixed":
            del data[0]["default_priority"]

        # If there is a validation error on the form, consume all status messages,
        # so that they don't appear in the form. We only want to show validation
        # messages directly on the respective field(s) in that case.
        if data[1]:
            status = IStatusMessage(self.request)
            status.show()
        return data



from Acquisition import aq_parent
from euphorie.content.browser import upload
from euphorie.content.module import Module
from euphorie.content.profilequestion import ProfileQuestion
from euphorie.content.risk import IFrenchEvaluation
from euphorie.content.risk import IKinneyEvaluation
from euphorie.content.risk import Risk
from euphorie.content.solution import Solution
from euphorie.testing import EuphorieFunctionalTestCase
from euphorie.testing import EuphorieIntegrationTestCase
from lxml import objectify
from plone import api
from plone.dexterity.utils import createContentInContainer
from plone.namedfile.file import NamedBlobImage
from plone.uuid.interfaces import IUUID


class SurveyImporterTests(EuphorieIntegrationTestCase):
    def setUp(self):
        super().setUp()
        self.client = self.portal.client
        self.loginAsPortalOwner()

    def createSurveyGroup(self):
        country = self.portal.sectors.nl
        sector = createContentInContainer(country, "euphorie.sector", title="sector")
        return createContentInContainer(sector, "euphorie.surveygroup", title="group")

    def createSurvey(self):
        surveygroup = self.createSurveyGroup()
        return createContentInContainer(surveygroup, "euphorie.survey", title="survey")

    def createModule(self):
        survey = self.createSurvey()
        return createContentInContainer(survey, "euphorie.module", title="module")

    def createRisk(self):
        module = self.createModule()
        return createContentInContainer(module, "euphorie.risk", title="risk")

    def testImportImage(self):
        snippet = objectify.fromstring(
            '<image caption="Key image" filename="myfile.gif" '
            'content-type="image/gif">R0lGODlhAQABAIAAAAAAAP//'
            "/yH5BAEAAAEALAAAAAABAAEAAAIBTAA7</image>"
        )
        importer = upload.SurveyImporter(None)
        (image, caption) = importer.ImportImage(snippet)
        self.assertEqual(caption, "Key image")
        self.assertTrue(isinstance(caption, str))
        self.assertEqual(image.contentType, "image/gif")
        self.assertEqual(image.filename, "myfile.gif")
        self.assertTrue(isinstance(image, NamedBlobImage))

    def testImportImage_filename_from_mimetype(self):
        snippet = objectify.fromstring(
            '<image caption="Key image" content-type="image/bmp">'
            "R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAEALAAAAAABAAEAAAIB"
            "TAA7</image>"
        )
        importer = upload.SurveyImporter(None)
        (image, caption) = importer.ImportImage(snippet)
        self.assertTrue(image.filename.endswith(".bmp"))

    def testImportImage_MimeFromFilename(self):
        snippet = objectify.fromstring(
            '<image filename="tiny.gif">R0lGODlhAQABAIAAAAAAAP///yH5BAEAAA'
            "EALAAAAAABAAEAAAIBTAA7</image>"
        )
        importer = upload.SurveyImporter(None)
        (image, caption) = importer.ImportImage(snippet)
        self.assertEqual(caption, None)
        self.assertEqual(image.filename, "tiny.gif")
        # Required by the interface :(
        self.assertTrue(isinstance(image.filename, str))
        self.assertEqual(image.contentType, "image/gif")

    def testImportSolution(self):
        snippet = objectify.fromstring(
            """<solution xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
             <description>Add more abstraction layers</description>
             <action-plan>Add another level</action-plan>
             <prevention-plan>Ask a code reviewer to verify the design</prevention-plan>
             <requirements>A good understanding of architecture</requirements>
           </solution>"""  # noqa
        )
        risk = self.createRisk()
        importer = upload.SurveyImporter(None)
        solution = importer.ImportSolution(snippet, risk)
        self.assertEqual(risk.keys(), ["3"])
        self.assertEqual(solution.description, "Add more abstraction layers")
        self.assertTrue(isinstance(solution.description, str))
        self.assertEqual(solution.action_plan, "Add another level")
        self.assertTrue(isinstance(solution.action_plan, str))
        self.assertEqual(
            solution.prevention_plan, "Ask a code reviewer to verify the design"
        )
        self.assertTrue(isinstance(solution.prevention_plan, str))
        self.assertEqual(solution.requirements, "A good understanding of architecture")
        self.assertTrue(isinstance(solution.requirements, str))

    def testImportSolutionFromEtranslate(self):
        snippet = objectify.fromstring(
            """<solution xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
             <description>Add more abstraction layers</description>
             <action>With <strong>HTML</strong><br/></action>
             <action-plan>Add another level</action-plan>
             <prevention-plan>Ask a code reviewer to verify the design</prevention-plan>
             <requirements>A good understanding of architecture</requirements>
           </solution>"""  # noqa
        )
        risk = self.createRisk()
        importer = upload.SurveyImporter(None)
        importer.is_etranslate_compatible = True
        solution = importer.ImportSolution(snippet, risk)
        self.assertEqual(risk.keys(), ["3"])
        self.assertEqual(solution.action, "With **HTML**  \n")
        self.assertTrue(isinstance(solution.action, str))

    def testImportSolution_MissingFields(self):
        snippet = objectify.fromstring(
            """<solution xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
             <description>Add more abstraction layers</description>
             <action-plan>Add another level</action-plan>
           </solution>"""
        )
        risk = self.createRisk()
        importer = upload.SurveyImporter(None)
        solution = importer.ImportSolution(snippet, risk)
        self.assertEqual(solution.description, "Add more abstraction layers")
        self.assertEqual(solution.action_plan, "Add another level")
        self.assertEqual(solution.prevention_plan, None)
        self.assertEqual(solution.requirements, None)

    def testImportRisk(self):
        snippet = objectify.fromstring(
            """<risk type="policy" xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
             <title>Are your desks at the right height?</title>
             <problem-description>Not all desks are set to the right height.</problem-description>
             <description>&lt;p&gt;The right height is important to prevent back problems.&lt;/p&gt;</description>
             <legal-reference>&lt;p&gt;See ARBO policies.&lt;/p&gt;</legal-reference>
             <show-not-applicable>true</show-not-applicable>
             <evaluation-method>direct</evaluation-method>
           </risk>"""  # noqa
        )
        module = self.createModule()
        importer = upload.SurveyImporter(None)
        risk = importer.ImportRisk(snippet, module)
        self.assertEqual(risk.title, "Are your desks at the right height?")
        self.assertTrue(isinstance(risk.title, str))
        self.assertEqual(
            risk.problem_description, "Not all desks are set to the right height."
        )
        self.assertTrue(isinstance(risk.problem_description, str))
        self.assertEqual(
            risk.description,
            "<p>The right height is important to prevent back problems.</p>",
        )
        self.assertTrue(isinstance(risk.description, str))
        self.assertEqual(risk.legal_reference, "<p>See ARBO policies.</p>")
        self.assertTrue(isinstance(risk.legal_reference, str))
        self.assertEqual(risk.show_notapplicable, True)
        self.assertEqual(risk.keys(), [])
        self.assertEqual(risk.caption, None)
        self.assertEqual(risk.image, None)

    def testImportRisk_Images(self):
        snippet = objectify.fromstring(
            """<risk type="policy" xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
             <title>Are your desks at the right height?</title>
             <problem-description>Not all desks are set to the right height.</problem-description>
             <description>&lt;p&gt;The right height is important to prevent back problems.&lt;/p&gt;</description>
             <legal-reference>&lt;p&gt;See ARBO policies.&lt;/p&gt;</legal-reference>
             <show-not-applicable>true</show-not-applicable>
             <evaluation-method>direct</evaluation-method>
             <image caption="Key image" filename="tiny.gif">R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAEALAAAAAABAAEAAAIBTAA7</image>
             <image caption="Image 2" filename="tiny.gif">R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAEALAAAAAABAAEAAAIBTAA7</image>
             <image caption="Image 3" filename="tiny.gif">R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAEALAAAAAABAAEAAAIBTAA7</image>
             <image caption="Image 4" filename="tiny.gif">R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAEALAAAAAABAAEAAAIBTAA7</image>
             <image caption="Image 5" filename="tiny.gif">R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAEALAAAAAABAAEAAAIBTAA7</image>
           </risk>"""  # noqa
        )
        module = self.createModule()
        importer = upload.SurveyImporter(None)
        risk = importer.ImportRisk(snippet, module)
        self.assertEqual(risk.image.filename, "tiny.gif")
        self.assertEqual(risk.caption, "Key image")
        self.assertEqual(risk.image2.filename, "tiny.gif")
        self.assertEqual(risk.caption2, "Image 2")
        self.assertEqual(risk.image3.filename, "tiny.gif")
        self.assertEqual(risk.caption3, "Image 3")
        self.assertEqual(risk.image4.filename, "tiny.gif")
        self.assertEqual(risk.caption4, "Image 4")

    def testImportRisk_DirectEvaluation(self):
        snippet = objectify.fromstring(
            """<risk type="risk" xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
             <title>Are your desks at the right height?</title>
             <description>&lt;p&gt;The right height is important to prevent back problems.&lt;/p&gt;</description>
             <evaluation-method default-priority="high">direct</evaluation-method>
           </risk>"""  # noqa
        )
        module = self.createModule()
        importer = upload.SurveyImporter(None)
        risk = importer.ImportRisk(snippet, module)
        self.assertEqual(risk.show_notapplicable, False)
        self.assertEqual(risk.evaluation_method, "direct")
        self.assertEqual(risk.default_priority, "high")

    def testImportRisk_CalculatedEvaluation_Kinney(self):
        snippet = objectify.fromstring(
            """<risk type="risk" xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
             <title>Are your desks at the right height?</title>
             <description>&lt;p&gt;The right height is important to prevent back problems.&lt;/p&gt;</description>
             <evaluation-method default-probability="small" default-frequency="regular" default-effect="high">calculated</evaluation-method>
           </risk>"""  # noqa
        )
        self.loginAsPortalOwner()
        module = self.createModule()
        importer = upload.SurveyImporter(None)
        importer.ImportRisk(snippet, module)
        risk = module["2"]
        self.assertTrue(IKinneyEvaluation.providedBy(risk))
        self.assertEqual(risk.show_notapplicable, False)
        self.assertEqual(risk.evaluation_method, "calculated")
        self.assertEqual(risk.default_probability, 1)
        self.assertEqual(risk.default_frequency, 4)
        self.assertEqual(risk.default_effect, 10)

    def testImportRisk_CalculatedEvaluation_French(self):
        snippet = objectify.fromstring(
            """<risk type="risk" xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
             <title>Are your desks at the right height?</title>
             <description>&lt;p&gt;The right height is important to prevent back problems.&lt;/p&gt;</description>
             <evaluation-method default-severity="very-severe" default-frequency="often">calculated</evaluation-method>
           </risk>"""  # noqa
        )
        self.loginAsPortalOwner()
        module = self.createModule()
        group = aq_parent(aq_parent(module))
        group.evaluation_algorithm = "french"
        importer = upload.SurveyImporter(None)
        risk = importer.ImportRisk(snippet, module)
        self.assertTrue(IFrenchEvaluation.providedBy(risk))
        self.assertEqual(risk.show_notapplicable, False)
        self.assertEqual(risk.evaluation_method, "calculated")
        self.assertEqual(risk.default_severity, 10)
        self.assertEqual(risk.default_frequency, 7)

    def testImportRisk_WithSolution(self):
        snippet = objectify.fromstring(
            """<risk type="policy" xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
             <title>Are your desks at the right height?</title>
             <description>&lt;p&gt;The right height is important to prevent back problems.&lt;/p&gt;</description>
             <evaluation-method>direct</evaluation-method>
             <solutions>
               <solution>
                 <description>Use height-adjustable desks</description>
                 <action-plan>Order height-adjustable desks for desk workers.</action-plan>
               </solution>
             </solutions>
           </risk>"""  # noqa
        )
        self.loginAsPortalOwner()
        module = self.createModule()
        importer = upload.SurveyImporter(None)
        risk = importer.ImportRisk(snippet, module)
        self.assertEqual(risk.keys(), ["3"])
        solution = risk["3"]
        self.assertTrue(isinstance(solution, Solution))
        self.assertEqual(solution.description, "Use height-adjustable desks")

    def testImportModule(self):
        snippet = objectify.fromstring(
            """<module optional="no" xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
             <title>Design patterns</title>
             <description>&lt;p&gt;Software design patterns are critical.&lt;/p&gt;</description>
             <solution-direction>&lt;p&gt;Buy the book from the gang of four.&lt;/p&gt;</solution-direction>
           </module>"""  # noqa
        )
        self.loginAsPortalOwner()
        survey = self.createSurvey()
        importer = upload.SurveyImporter(None)
        module = importer.ImportModule(snippet, survey)
        self.assertEqual(module.title, "Design patterns")
        self.assertTrue(isinstance(module.title, str))
        self.assertEqual(module.optional, False)
        self.assertEqual(module.question, None)
        self.assertEqual(
            module.description, "<p>Software design patterns are critical.</p>"
        )
        self.assertTrue(isinstance(module.description, str))
        self.assertEqual(
            module.solution_direction, "<p>Buy the book from the gang of four.</p>"
        )
        self.assertTrue(isinstance(module.solution_direction, str))
        self.assertEqual(module.keys(), [])
        self.assertEqual(module.image, None)

    def testImportModule_Image(self):
        snippet = objectify.fromstring(
            """<module optional="no" xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
             <title>Design patterns</title>
             <description>&lt;p&gt;Software design patterns are critical.&lt;/p&gt;</description>
             <solution-direction>&lt;p&gt;Buy the book from the gang of four.&lt;/p&gt;</solution-direction>
             <image caption="My caption" content-type="image/gif">R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAEALAAAAAABAAEAAAIBTAA7</image>
           </module>"""  # noqa
        )
        self.loginAsPortalOwner()
        survey = self.createSurvey()
        importer = upload.SurveyImporter(None)
        module = importer.ImportModule(snippet, survey)
        self.assertNotEqual(module.image, None)
        self.assertEqual(module.image.contentType, "image/gif")
        self.assertEqual(module.caption, "My caption")

    def testImportModule_Optional(self):
        snippet = objectify.fromstring(
            """<module optional="true" xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
             <title>Design patterns</title>
             <description>&lt;p&gt;Software design patterns are critical.&lt;/p&gt;</description>
             <question>Have you used design patterns?</question>
           </module>"""  # noqa
        )
        self.loginAsPortalOwner()
        survey = self.createSurvey()
        importer = upload.SurveyImporter(None)
        module = importer.ImportModule(snippet, survey)
        self.assertEqual(module.optional, True)
        self.assertEqual(module.question, "Have you used design patterns?")

    def testImportModule_WithSubModule(self):
        snippet = objectify.fromstring(
            """<module optional="no" xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
             <title>Design patterns</title>
             <description>&lt;p&gt;Software design patterns are critical.&lt;/p&gt;</description>
             <module optional="no">
               <title>Iterators</title>
               <description>&lt;p&gt;Iterators help optimise list handling.&lt;/p&gt;</description>
             </module>
           </module>"""  # noqa
        )
        self.loginAsPortalOwner()
        survey = self.createSurvey()
        importer = upload.SurveyImporter(None)
        module = importer.ImportModule(snippet, survey)
        self.assertEqual(module.keys(), ["2"])
        module = module["2"]
        self.assertTrue(isinstance(module, Module))
        self.assertEqual(module.title, "Iterators")

    def testImportModule_WithRisk(self):
        snippet = objectify.fromstring(
            """<module optional="no" xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
             <title>Design patterns</title>
             <description>&lt;p&gt;Software design patterns are critical.&lt;/p&gt;</description>
             <risk type="policy">
               <title>New hires are not aware of design patterns.</title>
               <description>&lt;p&gt;Every developer should know about them..&lt;/p&gt;</description>
               <evaluation-method>direct</evaluation-method>
             </risk>
           </module>"""  # noqa
        )
        self.loginAsPortalOwner()
        survey = self.createSurvey()
        importer = upload.SurveyImporter(None)
        module = importer.ImportModule(snippet, survey)
        self.assertEqual(module.keys(), ["2"])
        risk = module["2"]
        self.assertTrue(isinstance(risk, Risk))
        self.assertEqual(risk.title, "New hires are not aware of design patterns.")

    def testImportProfileQuestion(self):
        snippet = objectify.fromstring(
            """<profile-question xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
             <title>Laptop usage</title>
             <question>Do your employees use laptops?</question>
             <description>&lt;p&gt;Laptops are very common in the modern workplace.&lt;/p&gt;</description>
           </profile-question>"""  # noqa
        )
        self.loginAsPortalOwner()
        survey = self.createSurvey()
        importer = upload.SurveyImporter(None)
        profile = importer.ImportProfileQuestion(snippet, survey)
        self.assertEqual(profile.title, "Laptop usage")
        self.assertTrue(isinstance(profile.title, str))
        self.assertEqual(profile.question, "Do your employees use laptops?")
        self.assertTrue(isinstance(profile.question, str))
        self.assertEqual(
            profile.description,
            "<p>Laptops are very common in the modern workplace.</p>",
        )
        self.assertTrue(isinstance(profile.description, str))
        self.assertEqual(profile.keys(), [])

    def testImportProfileQuestion_optional_type(self):
        snippet = objectify.fromstring(
            """<profile-question type="optional"
                xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
             <title>Laptop usage</title>
             <question>Do your employees use laptops?</question>
           </profile-question>"""
        )
        self.loginAsPortalOwner()
        survey = self.createSurvey()
        importer = upload.SurveyImporter(None)
        profile = importer.ImportProfileQuestion(snippet, survey)
        self.assertTrue(isinstance(profile, Module))
        self.assertEqual(profile.optional, True)

    def testImportProfileQuestion_WithModule(self):
        snippet = objectify.fromstring(
            """<profile-question xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
             <title>Laptop usage</title>
             <question>Do your employees use laptops?</question>
             <description>&lt;p&gt;Laptops are very common in the modern workplace.&lt;/p&gt;</description>
             <module optional="no">
               <title>Design patterns</title>
               <description>&lt;p&gt;Software design patterns are critical.&lt;/p&gt;</description>
             </module>
           </profile-question>"""  # noqa
        )
        self.loginAsPortalOwner()
        survey = self.createSurvey()
        importer = upload.SurveyImporter(None)
        profile = importer.ImportProfileQuestion(snippet, survey)
        self.assertEqual(profile.keys(), ["2"])
        module = profile["2"]
        self.assertTrue(isinstance(module, Module))
        self.assertEqual(module.title, "Design patterns")

    def testImportProfileQuestion_WithRisk(self):
        snippet = objectify.fromstring(
            """<profile-question xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
             <title>Laptop usage</title>
             <question>Do your employees use laptops?</question>
             <description>&lt;p&gt;Laptops are very common in the modern workplace.&lt;/p&gt;</description>
             <risk type="policy">
               <title>New hires are not aware of design patterns.</title>
               <description>&lt;p&gt;Every developer should know about them..&lt;/p&gt;</description>
               <evaluation-method>direct</evaluation-method>
             </risk>
           </profile-question>"""  # noqa
        )
        self.loginAsPortalOwner()
        survey = self.createSurvey()
        importer = upload.SurveyImporter(None)
        profile = importer.ImportProfileQuestion(snippet, survey)
        self.assertEqual(profile.keys(), ["2"])
        risk = profile["2"]
        self.assertTrue(isinstance(risk, Risk))
        self.assertEqual(risk.title, "New hires are not aware of design patterns.")

    def testImportSurvey(self):
        snippet = objectify.fromstring(
            """<survey xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
             <title>Software development</title>
             <classification-code>A.1.2.3</classification-code>
             <language>nl</language>
             <evaluation-optional>true</evaluation-optional>
           </survey>"""
        )
        self.loginAsPortalOwner()
        surveygroup = self.createSurveyGroup()
        importer = upload.SurveyImporter(None)
        survey = importer.ImportSurvey(snippet, surveygroup, "Fresh import")
        self.assertEqual(surveygroup.keys(), ["fresh-import"])
        self.assertEqual(survey.keys(), [])
        self.assertEqual(survey.title, "Fresh import")
        self.assertTrue(isinstance(survey.title, str))
        self.assertEqual(survey.classification_code, "A.1.2.3")
        self.assertTrue(isinstance(survey.language, str))
        self.assertEqual(survey.language, "nl")
        self.assertTrue(isinstance(survey.classification_code, str))
        self.assertEqual(survey.evaluation_optional, True)

    def testImportSurveyFromEtranslate(self):
        snippet = objectify.fromstring(
            """<survey xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
             <title>Software development</title>
             <measures_text_handling value="partial"/>
             <tool_type value="existing_measures"/>
             <evaluation-optional value="true"/>
             <integrated_action_plan>false</integrated_action_plan>
           </survey>"""
        )
        self.loginAsPortalOwner()
        surveygroup = self.createSurveyGroup()
        importer = upload.SurveyImporter(None)
        importer.is_etranslate_compatible = True
        survey = importer.ImportSurvey(snippet, surveygroup, "Fresh import")
        self.assertEqual(survey.measures_text_handling, "partial")
        self.assertEqual(survey.tool_type, "existing_measures")
        self.assertEqual(survey.evaluation_optional, True)
        self.assertEqual(survey.integrated_action_plan, False)

    def testImportSurvey_WithModule(self):
        snippet = objectify.fromstring(
            """<survey optional="no" xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
             <title>Software development</title>
             <module optional="no">
               <title>Design patterns</title>
               <description>&lt;p&gt;Software design patterns are critical.&lt;/p&gt;</description>
             </module>
           </survey>"""  # noqa
        )
        self.loginAsPortalOwner()
        surveygroup = self.createSurveyGroup()
        importer = upload.SurveyImporter(None)
        survey = importer.ImportSurvey(snippet, surveygroup, "Fresh import")
        self.assertEqual(surveygroup.keys(), ["fresh-import"])
        self.assertEqual(survey.keys(), ["1"])
        module = survey["1"]
        self.assertTrue(isinstance(module, Module))
        self.assertEqual(module.title, "Design patterns")

    def testImportSurvey_WithProfileQuestion(self):
        snippet = objectify.fromstring(
            """<survey optional="no" xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
             <title>Software development</title>
             <profile-question>
               <title>Laptop usage</title>
               <question>Do your employees use laptops?</question>
               <description>&lt;p&gt;Laptops are very common in the modern workplace.&lt;/p&gt;</description>
             </profile-question>
           </survey>"""  # noqa
        )
        self.loginAsPortalOwner()
        surveygroup = self.createSurveyGroup()
        importer = upload.SurveyImporter(None)
        survey = importer.ImportSurvey(snippet, surveygroup, "Fresh import")
        self.assertEqual(surveygroup.keys(), ["fresh-import"])
        self.assertEqual(survey.keys(), ["1"])
        profile = survey["1"]
        self.assertTrue(isinstance(profile, ProfileQuestion))
        self.assertEqual(profile.title, "Laptop usage")

    def testImportSurvey_ChildOrdering(self):
        snippet = objectify.fromstring(
            """<survey optional="no" xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
             <title>Software development</title>
             <module optional="no">
               <title>Module one</title>
               <description/>
             </module>
             <profile-question>
               <title>Profile one</title>
               <question/>
               <description/>
             </profile-question>
             <module optional="no">
               <title>Module two</title>
               <description/>
             </module>
             <profile-question>
               <title>Profile two</title>
               <question/>
               <description/>
             </profile-question>
           </survey>"""
        )
        self.loginAsPortalOwner()
        surveygroup = self.createSurveyGroup()
        importer = upload.SurveyImporter(None)
        survey = importer.ImportSurvey(snippet, surveygroup, "Fresh import")
        self.assertEqual(surveygroup.keys(), ["fresh-import"])
        self.assertEqual(survey.keys(), ["1", "2", "3", "4"])
        children = survey.values()
        self.assertEqual(children[0].title, "Module one")
        self.assertEqual(children[1].title, "Profile one")
        self.assertEqual(children[2].title, "Module two")
        self.assertEqual(children[3].title, "Profile two")

    def testImportTrainingQuestion(self):
        snippet = objectify.fromstring(
            """<training_question>
              <title>How many rivets were used in the Titanic?</title>
              <right_answer>3,000,000</right_answer>
              <wrong_answer_1>None, it was glued together</wrong_answer_1>
              <wrong_answer_2>300,000</wrong_answer_2>
            </training_question>
            """
        )
        self.loginAsPortalOwner()
        survey = self.createSurvey()
        survey.enable_web_training = True
        survey.enable_test_questions = True
        importer = upload.SurveyImporter(None)
        api.portal.set_registry_record("euphorie.use_training_module", True)
        training_question = importer.ImportTrainingQuestion(snippet, survey)
        self.assertEqual(
            training_question.title, "How many rivets were used in the Titanic?"
        )
        self.assertEqual(training_question.right_answer, "3,000,000")
        self.assertEqual(
            training_question.wrong_answer_1, "None, it was glued together"
        )
        self.assertEqual(training_question.wrong_answer_2, "300,000")


class SectorImporterTests(EuphorieFunctionalTestCase):
    def testImportSurvey(self):
        snippet = objectify.fromstring(
            """<sector xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
             <title>Software development sector</title>
             <survey>
               <title>Software development</title>
             </survey>
           </sector>"""
        )
        self.loginAsPortalOwner()
        country = self.portal.sectors.nl
        importer = upload.SectorImporter(country)
        importer(snippet, None, "login", None, "Import")
        self.assertEqual(country.keys(), ["help", "software-development-sector"])
        sector = country["software-development-sector"]
        self.assertEqual(sector.title, "Software development sector")
        self.assertEqual(sector.login, "login")
        self.assertTrue(isinstance(sector.title, str))
        self.assertEqual(sector.keys(), ["software-development"])
        self.assertEqual(sector.contact_email, None)
        self.assertEqual(sector.password, None)
        group = sector["software-development"]
        self.assertEqual(group.title, "Software development")
        self.assertTrue(isinstance(group.title, str))
        self.assertEqual(group.keys(), ["import"])
        self.assertTrue(isinstance(IUUID(sector), str))

    def testImportSurvey_WithAccount(self):
        snippet = objectify.fromstring(
            """<sector xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
             <title>Software development sector</title>
             <account login="coders" password="s3cr3t" />
             <contact>
               <name>Team lead</name>
               <email>discard@simplon.biz</email>
             </contact>
           </sector>"""
        )
        self.loginAsPortalOwner()
        country = self.portal.sectors.nl
        importer = upload.SectorImporter(country)
        importer(snippet, None, None, None, "Import")
        sector = country["software-development-sector"]
        self.assertEqual(sector.login, "coders")
        self.assertEqual(sector.contact_name, "Team lead")
        self.assertEqual(sector.contact_email, "discard@simplon.biz")
        self.assertEqual(sector.password, "s3cr3t")

    def testImportSurvey_WithLogo(self):
        snippet = objectify.fromstring(
            """<sector xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
             <title>Software development sector</title>
             <logo filename="tiny.gif">R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAEALAAAAAABAAEAAAIBTAA7</logo>
           </sector>"""  # noqa
        )
        self.loginAsPortalOwner()
        country = self.portal.sectors.nl
        importer = upload.SectorImporter(country)
        importer(snippet, None, "login", None, "Import")
        sector = country["software-development-sector"]
        self.assertNotEqual(sector.logo, None)
        self.assertEqual(sector.logo.filename, "tiny.gif")
        self.assertEqual(sector.logo.contentType, "image/gif")
        self.assertEqual(
            sector.logo.data,
            b"GIF89a\x01\x00\x01\x00\x80\x00\x00"
            b"\x00\x00\x00\xff\xff\xff!\xf9\x04"
            b"\x01\x00\x00\x01\x00,\x00\x00\x00"
            b"\x00\x01\x00\x01\x00\x00\x02\x01L"
            b"\x00;",
        )



"""Empty message.

Revision ID: 27
Revises: 26
Create Date: 2020-03-04 10:51:17.559416
"""

from alembic import op
from euphorie.deployment.upgrade.utils import has_column

import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "27"
down_revision = "26"
branch_labels = None
depends_on = None


def upgrade():
    if not has_column("action_plan", "action"):
        op.add_column(
            "action_plan", sa.Column("action", sa.UnicodeText(), nullable=True)
        )
    if not has_column("action_plan", "solution_id"):
        op.add_column(
            "action_plan", sa.Column("solution_id", sa.String(length=20), nullable=True)
        )
    if not has_column("action_plan", "plan_type"):
        op.add_column(
            "action_plan", sa.Column("plan_type", sa.String(length=20), nullable=True)
        )
        op.create_index(
            op.f("ix_action_plan_plan_type"), "action_plan", ["plan_type"], unique=False
        )
        op.execute("UPDATE action_plan SET plan_type = 'measure_custom'")
        op.execute(
            """
UPDATE action_plan ap
    SET "action" = CASE
    WHEN ap.prevention_plan is not NULL
    THEN ap.action_plan || E'\n' || ap.prevention_plan
    ELSE ap.action_plan END"""
        )
        op.alter_column("action_plan", "plan_type", nullable=False)

    if not has_column("session", "migrated"):
        op.add_column("session", sa.Column("migrated", sa.DateTime(), nullable=True))


def downgrade():
    op.drop_index(op.f("ix_action_plan_plan_type"), table_name="action_plan")
    op.drop_column("action_plan", "solution_id")
    op.drop_column("action_plan", "plan_type")
    op.drop_column("action_plan", "action")
    op.drop_column("session", "migrated")



"""Add session redirect

Revision ID: 20230406070728
Revises: 20220720110538
Create Date: 2023-04-06 07:15:08.783635

"""

from alembic import op
from euphorie.deployment.upgrade.utils import has_table

import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "20230406070728"
down_revision = "20220720110538"
branch_labels = None
depends_on = None


def upgrade():
    if not has_table("session_redirect"):
        op.create_table(
            "session_redirect",
            sa.Column("old_session_id", sa.Integer(), nullable=False),
            sa.Column("new_session_id", sa.Integer(), nullable=False),
            sa.PrimaryKeyConstraint("old_session_id"),
        )


def downgrade():
    if has_table("session_redirect"):
        op.drop_table("session_redirect")



from ftw.upgrade import UpgradeStep


class AddUseConsultancyPhaseFeatureSwitch(UpgradeStep):
    """Add use_consultancy_phase feature switch."""

    def __call__(self):
        self.install_upgrade_profile()



"""
Sector
------

The client proxy for displaying sector information.
"""

from Acquisition import aq_base
from Acquisition import aq_parent
from plone.app.dexterity.behaviors.metadata import IBasic
from plone.dexterity.content import Container
from plone.supermodel import model
from Products.CMFCore.interfaces import ISiteRoot
from zope.component import getUtility
from zope.interface import implementer


class IClientSector(model.Schema, IBasic):
    """Sector information in the online client.

    This object proxies the title and logo from the main sector object.

    The online client is implemented as a cantainer with all available surveys.
    The default view for all survey elements inside this container is changed
    to the client user interface. This is done using a simple traversal
    adapter.
    """


@implementer(IClientSector)
class ClientSector(Container):
    def _sector(self):
        sectors = getUtility(ISiteRoot).sectors
        country = getattr(sectors, aq_parent(self).id)
        return getattr(country, self.id)

    def Title(self):
        title = getattr(aq_base(self), "title", None)
        if title is not None:
            return title
        else:
            return self._sector().title



"""
Profile Question
----------------

A Profile Question is a container for Modules. A question is used to determine
whether or not a Module should be enabled, or whether it should be repeated.

portal_type: euphorie.profilequestion
"""

from .. import MessageFactory as _
from .behaviour.richdescription import IRichDescription
from .behaviour.uniqueid import get_next_id
from .behaviour.uniqueid import INameFromUniqueId
from .fti import check_fti_paste_allowed
from .interfaces import IQuestionContainer
from .module import ConstructionFilter
from .module import item_depth
from .module import tree_depth
from .utils import StripMarkup
from euphorie.content.dependency import ConditionalTextLine
from plone.app.dexterity.behaviors.metadata import IBasic
from plone.autoform import directives
from plone.dexterity.content import Container
from plone.indexer import indexer
from plone.supermodel import model
from plonetheme.nuplone.z3cform.directives import depends
from zope import schema
from zope.interface import implementer

import sys


class IProfileQuestion(model.Schema, IRichDescription, IBasic):
    """Survey Profile question.

    A profile question is used to determine if parts of a survey should
    be skipped, or repeated multiple times.
    """

    question = schema.TextLine(
        title=_("label_profilequestion_question", default="Question"),
        description=_(
            "This question must ask the user if this profile " "applies to them."
        ),
        required=True,
    )
    directives.widget(question="euphorie.content.risk.TextLines4Rows")

    use_location_question = schema.Bool(
        title=_(
            "label_use_location_question",
            default="Ask the user about (multiple) locations?",
        ),
        description=_(
            "description_use_location_question",
            default="If this part is active, the user will be asked to "
            "enter the name of all locations for which this module applies. "
            "This means, the module will be repeated as many times as there "
            "are locations. If you do not need this repeatable behaviour, "
            "untick the checkbox to turn it off.",
        ),
        required=False,
        default=True,
    )

    depends("label_multiple_present", "use_location_question", "on")
    label_multiple_present = ConditionalTextLine(
        title=_("Multiple item question"),
        required=True,
        description=_(
            "This question must ask the user if the service is "
            "offered in more than one location."
        ),
    )
    directives.widget(label_multiple_present="euphorie.content.risk.TextLines4Rows")

    depends("label_single_occurance", "use_location_question", "on")
    label_single_occurance = ConditionalTextLine(
        title=_("Single occurance prompt"),
        description=_(
            "This must ask the user for the name of the " "relevant location."
        ),
        required=True,
    )
    directives.widget(label_single_occurance="euphorie.content.risk.TextLines4Rows")

    depends("label_multiple_occurances", "use_location_question", "on")
    label_multiple_occurances = ConditionalTextLine(
        title=_("Multiple occurance prompt"),
        description=_(
            "This must ask the user for the names of all " "relevant locations."
        ),
        required=True,
    )
    directives.widget(label_multiple_occurances="euphorie.content.risk.TextLines4Rows")


@implementer(IProfileQuestion, IQuestionContainer)
class ProfileQuestion(Container):
    portal_type = "euphorie.profilequestion"

    question = None
    image = None
    optional = False

    def _get_id(self, orig_id):
        """Pick an id for pasted content."""
        frame = sys._getframe(1)
        ob = frame.f_locals.get("ob")
        if ob is not None and INameFromUniqueId.providedBy(ob):
            return get_next_id(self)
        return super()._get_id(orig_id)

    def _verifyObjectPaste(self, object, validate_src=True):
        super()._verifyObjectPaste(object, validate_src)
        if validate_src:
            check_fti_paste_allowed(self, object)
            if IQuestionContainer.providedBy(object):
                my_depth = item_depth(self)
                paste_depth = tree_depth(object)
                if my_depth + paste_depth > ConstructionFilter.maxdepth:
                    raise ValueError("Pasting would create a too deep structure.")


@indexer(IProfileQuestion)
def SearchableTextIndexer(obj):
    """Index the title and description."""
    return " ".join([obj.title, StripMarkup(obj.description)])



"""
Settings
--------

Change a user's password/email or delete an account.
"""

from Acquisition import aq_inner
from euphorie.client import MessageFactory as _
from euphorie.client.interfaces import INotificationCategory
from euphorie.client.model import Account
from euphorie.client.model import AccountChangeRequest
from euphorie.client.model import get_current_account
from euphorie.client.model import NotificationSubscription
from euphorie.client.utils import CreateEmailTo
from euphorie.client.utils import randomString
from plone import api
from plone.autoform import directives
from plone.autoform.form import AutoExtensibleForm
from plone.memoize.instance import memoize
from plone.protect.interfaces import IDisableCSRFProtection
from plone.schema import Email
from plone.supermodel import model
from Products.CMFCore.utils import getToolByName
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from Products.MailHost.MailHost import MailHostError
from Products.statusmessages.interfaces import IStatusMessage
from urllib.parse import urlencode
from z3c.form import button
from z3c.form import form
from z3c.form.interfaces import WidgetActionExecutionError
from z3c.saconfig import Session
from zope import schema
from zope.component import getAdapters
from zope.i18n import translate
from zope.interface import alsoProvides
from zope.interface import directlyProvides
from zope.interface import Invalid
from zope.interface import invariant

import datetime
import logging
import smtplib


log = logging.getLogger(__name__)


class UserMenu(BrowserView):
    @property
    @memoize
    def webhelpers(self):
        return api.content.get_view("webhelpers", self.context, self.request)

    @property
    def menu_contents(self):
        return [
            {
                "class_outer": "menu-item-preferences",
                "class_inner": "icon-cog",
                "url": f"{self.webhelpers.country_url}/preferences",
                "label": _("title_preferences", default="Preferences"),
            },
            {
                "class_outer": "menu-item-change-password",
                "class_inner": "icon-key",
                "url": f"{self.webhelpers.country_url}/account-settings",
                "label": _("title_change_password", default="Change password"),
            },
            {
                "class_outer": "menu-item-change-email-address",
                "class_inner": "icon-mail",
                "url": f"{self.webhelpers.country_url}/new-email",
                "label": _("Change email address"),
            },
            {
                "class_outer": "menu-item-delete-account",
                "class_inner": "icon-cancel-circle",
                "url": f"{self.webhelpers.country_url}/account-delete",
                "label": _("Delete account"),
            },
            {
                "class_outer": "menu-item-logout",
                "class_inner": "icon-exit",
                "url": f"{self.webhelpers.country_url}/logout",
                "label": _("navigation_logout", default="Logout"),
            },
        ]


class PasswordChangeSchema(model.Schema):
    old_password = schema.Password(
        title=_("label_old_password", default="Current Password"),
        required=True,
    )
    directives.widget(old_password="z3c.form.browser.password.PasswordFieldWidget")

    new_password = schema.Password(
        title=_("label_new_password", default="Desired password")
    )
    directives.widget(new_password="z3c.form.browser.password.PasswordFieldWidget")

    new_password_confirmation = schema.Password(
        title=_("label_new_password_confirmation", default="Again password")
    )
    directives.widget(
        new_password_confirmation="z3c.form.browser.password.PasswordFieldWidget"
    )

    @invariant
    def validate_same_value(data):
        if data.new_password != data.new_password_confirmation:
            raise Invalid(_("Password doesn't compare with confirmation value"))


class AccountDeleteSchema(model.Schema):
    password = schema.Password(title=_("Your password for confirmation"), required=True)
    directives.widget(password="z3c.form.browser.password.PasswordFieldWidget")


class EmailChangeSchema(model.Schema):
    loginname = Email(title=_("Email address/account name"), required=True)
    directives.widget("loginname", type="email")

    password = schema.Password(title=_("Your password for confirmation"), required=True)
    directives.widget(password="z3c.form.browser.password.PasswordFieldWidget")


class PreferencesSchema(model.Schema):
    first_name = schema.TextLine(
        title=_("label_first_name", default="First name"), required=False
    )
    last_name = schema.TextLine(
        title=_("label_last_name", default="Last name"), required=False
    )


class Preferences(AutoExtensibleForm, form.Form):
    template = ViewPageTemplateFile("templates/preferences.pt")
    no_splash = True
    schema = PreferencesSchema

    @property
    def show_personal_details(self):
        return api.portal.get_registry_record(
            "euphorie.personal_details__enabled", default=True
        )

    @property
    def show_notifications(self):
        return api.portal.get_registry_record(
            "euphorie.notifications__enabled", default=False
        )

    @property
    def allow_notification_settings(self):
        return api.portal.get_registry_record(
            "euphorie.notifications__allow_user_settings", default=True
        )

    @property
    @memoize
    def current_user(self):
        return get_current_account()

    def getContent(self):
        user = self.current_user
        directlyProvides(user, PreferencesSchema)
        return user

    @property
    def all_notifications(self):
        notifications = getAdapters((self.context, self.request), INotificationCategory)
        # Return the available notifications, sorted by title.
        notifications = [
            notification[1]
            for notification in notifications
            if notification[1].available
        ]
        return sorted(
            notifications,
            key=lambda notification: translate(notification.title),
        )

    @property
    @memoize
    def existing_notification_subscriptions(self):
        subscriptions = Session.query(NotificationSubscription).filter(
            NotificationSubscription.account_id == self.current_user.getId()
        )
        return {subscription.category: subscription for subscription in subscriptions}

    def subscribe_notification(self, category_id):
        notification = self.existing_notification_subscriptions.get(category_id)
        if notification:
            notification.enabled = True
            return
        Session.add(
            NotificationSubscription(
                account_id=self.current_user.getId(),
                category=category_id,
                enabled=True,
            )
        )

    def unsubscribe_notification(self, category_id):
        notification = self.existing_notification_subscriptions.get(category_id)
        if notification:
            notification.enabled = False
            return
        Session.add(
            NotificationSubscription(
                account_id=self.current_user.getId(),
                category=category_id,
                enabled=False,
            )
        )

    @button.buttonAndHandler(_("Save"), name="save")
    def handleSave(self, action):
        flash = IStatusMessage(self.request).addStatusMessage
        (data, errors) = self.extractData()
        if errors:
            for error in errors:
                flash(error.message, "notice")
            return
        user = self.current_user
        user.first_name = data["first_name"]
        user.last_name = data["last_name"]

        if self.show_notifications and self.allow_notification_settings:
            for notification in self.all_notifications:
                if self.request.get("notifications", {}).get(notification.id):
                    self.subscribe_notification(notification.id)
                else:
                    self.unsubscribe_notification(notification.id)


class AccountSettings(AutoExtensibleForm, form.Form):
    """View name: @@account-settings."""

    template = ViewPageTemplateFile("templates/account-settings.pt")

    schema = PasswordChangeSchema
    ignoreContext = True

    label = _("title_change_password", default="Change password")

    def updateWidgets(self):
        super().updateWidgets()
        self.widgets["old_password"].addClass("password")
        self.widgets["new_password"].addClass("password")

    @button.buttonAndHandler(_("Save changes"), name="save")
    def handleSave(self, action):
        flash = IStatusMessage(self.request).addStatusMessage
        (data, errors) = self.extractData()
        if errors:
            for error in errors:
                flash(error.message, "notice")
            return

        user = get_current_account()
        if not data["new_password"]:
            flash(_("There were no changes to be saved."), "notice")
            return
        login_view = api.content.get_view(
            name="login",
            context=self.context,
            request=self.request,
        )
        error = login_view.check_password_policy(data["new_password"])
        if error:
            raise WidgetActionExecutionError("new_password", Invalid(error))
        if not user.verify_password(data["old_password"]):
            raise WidgetActionExecutionError(
                "old_password", Invalid(_("Invalid password"))
            )
        user.password = data["new_password"]
        flash(_("Your password was successfully changed."), "success")

    @button.buttonAndHandler(_("button_cancel", default="Cancel"), name="cancel")
    def handleCancel(self, action):
        self.request.response.redirect(self.request.client.absolute_url())


class DeleteAccount(AutoExtensibleForm, form.Form):
    """View name: @@account-delete."""

    template = ViewPageTemplateFile("templates/account-delete.pt")

    schema = AccountDeleteSchema
    ignoreContext = True

    label = _("title_account_delete", default="Delete account")

    def updateWidgets(self):
        super().updateWidgets()
        self.widgets["password"].addClass("password")

    def logout(self):
        pas = getToolByName(self.context, "acl_users")
        pas.resetCredentials(self.request, self.request.response)

    @button.buttonAndHandler(_("Delete account"), name="delete")
    def handleDelete(self, action):
        (data, errors) = self.extractData()
        if errors:
            return

        user = get_current_account()
        if not user.verify_password(data["password"]):
            raise WidgetActionExecutionError("password", Invalid(_("Invalid password")))

        Session.delete(user)
        self.logout()
        self.request.response.redirect(self.request.client.absolute_url())

    @button.buttonAndHandler(_("button_cancel", default="Cancel"), name="cancel")
    def handleCancel(self, action):
        self.request.response.redirect(self.request.client.absolute_url())


class NewEmail(AutoExtensibleForm, form.Form):
    template = ViewPageTemplateFile("templates/new-email.pt")

    schema = EmailChangeSchema
    email_template = ViewPageTemplateFile("templates/confirm-email.pt")
    ignoreContext = True

    label = _("title_change_email", default="Change email address")

    @property
    def email_from_name(self):
        return api.portal.get_registry_record("plone.email_from_name")

    @property
    def email_from_address(self):
        return api.portal.get_registry_record("plone.email_from_address")

    def updateFields(self):
        super().updateFields()
        self.fields["password"].ignoreContext = True

    def updateWidgets(self):
        super().updateWidgets()
        self.widgets["password"].addClass("password")

    def getContent(self):
        user = get_current_account()
        directlyProvides(user, EmailChangeSchema)
        return user

    def initiateRequest(self, account, login):
        flash = IStatusMessage(self.request).addStatusMessage
        # Make it work when acl_users is in Memcached: We need to fetch the
        # account again, to prevent DetachedInstanceError
        account_query = Session.query(Account).filter(Account.id == account.id)
        if not account_query.count():
            log.error("Account could not be fetched")
            flash(
                _("An error occured while sending the confirmation email."),
                "error",
            )
            return False
        account = account_query.one()
        if account.change_request is None:
            account.change_request = AccountChangeRequest()
        account.change_request.id = randomString()
        account.change_request.expires = datetime.datetime.now() + datetime.timedelta(
            days=7
        )
        account.change_request.value = login

        client_url = self.request.client.absolute_url()
        confirm_url = "{}/confirm-change?{}".format(
            client_url,
            urlencode({"key": account.change_request.id}),
        )

        mailhost = getToolByName(self.context, "MailHost")
        body = self.email_template(
            account=account,
            new_login=login,
            client_url=client_url,
            confirm_url=confirm_url,
        )
        subject = translate(
            _("Confirm OiRA email address change"), context=self.request
        )
        mail = CreateEmailTo(
            self.email_from_name, self.email_from_address, login, subject, body
        )

        try:
            mailhost.send(mail, login, self.email_from_address, immediate=True)
            log.info("Sent email confirmation to %s", account.email)
        except MailHostError as e:
            log.error(
                "MailHost error sending email confirmation to %s: %s",
                account.email,
                e,
            )
            flash(
                _("An error occured while sending the confirmation email."),
                "error",
            )
            return False
        except smtplib.SMTPException as e:
            log.error(
                "smtplib error sending the confirmation email to %s: %s",
                account.email,
                e,
            )
            flash(
                _("An error occured while sending the confirmation email."),
                "error",
            )
            return False
        except OSError as e:
            log.error(
                "Socket error sending confirmation email to %s: %s",
                account.email,
                e[1],
            )
            flash(
                _("An error occured while sending the confirmation email."),
                "error",
            )
            return False

        return True

    @button.buttonAndHandler(_("Save changes"), name="save")
    def handleSave(self, action):
        flash = IStatusMessage(self.request).addStatusMessage

        (data, errors) = self.extractData()
        if errors:
            return
        url = self.context.absolute_url()

        user = get_current_account()
        if not user.verify_password(data["password"]):
            raise WidgetActionExecutionError("password", Invalid(_("Invalid password")))

        settings_url = "%s/account-settings" % url
        if not data["loginname"] or data["loginname"].strip() == user.loginname:
            self.request.response.redirect(settings_url)
            flash(_("There were no changes to be saved."), "notice")
            return

        login = data["loginname"].strip().lower()
        existing = Session.query(Account.id).filter(Account.loginname == login)
        if existing.count():
            raise WidgetActionExecutionError(
                "loginname", Invalid(_("This email address is not available."))
            )

        self.initiateRequest(user, login)

        flash(
            _(
                "email_change_pending",
                default=(
                    "Please confirm your new email address by clicking on "
                    "the link in the email that will be sent in a few "
                    'minutes to "${email}". Please note that the new '
                    "email address is also your new login name."
                ),
                mapping={"email": data["loginname"]},
            ),
            "warning",
        )
        self.request.response.redirect("%s/" % url)

    @button.buttonAndHandler(_("button_cancel", default="Cancel"), name="cancel")
    def handleCancel(self, action):
        self.request.response.redirect(self.request.client.absolute_url())


class ChangeEmail(BrowserView):
    def __call__(self):
        url = "%s/" % aq_inner(self.context).absolute_url()
        flash = IStatusMessage(self.request).addStatusMessage
        key = self.request.get("key")
        if key is None:
            flash(_("This request could not be processed."), "warning")
            self.request.response.redirect(url)
            return

        request = Session.query(AccountChangeRequest).get(key)
        if request is None:
            flash(_("This request could not be processed."), "warning")
            self.request.response.redirect(url)
            return

        request.account.loginname = request.value
        alsoProvides(self.request, IDisableCSRFProtection)
        Session.delete(request)
        flash(_("Your email address has been updated."), "success")
        self.request.response.redirect(url)



"""Empty message.

Revision ID: 26
Revises: 25
Create Date: 2020-02-03 15:42:46.758366
"""

from alembic import op
from euphorie.deployment.upgrade.utils import has_column

import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "26"
down_revision = "25"
branch_labels = None
depends_on = None


def upgrade():
    if not has_column("session", "completion_percentage"):
        op.add_column(
            "session", sa.Column("completion_percentage", sa.Integer(), nullable=True)
        )


def downgrade():
    op.drop_column("session", "completion_percentage")



"""
Documentation
-------------

A folder that contains documentation for a language.

https://admin.oiraproject.eu/documents/en

portal_type: euphorie.documentation
"""

from plone.app.dexterity.behaviors.metadata import IBasic
from plone.supermodel import model


class IDocumentationFolder(model.Schema, IBasic):
    """A folder for all documentation for a specific language."""



"""Empty message.

Revision ID: 25
Revises: 24
Create Date: 2019-12-04 11:58:00.148524
"""

from alembic import op
from euphorie.deployment.upgrade.utils import has_column

import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "25"
down_revision = "24"
branch_labels = None
depends_on = None


def upgrade():
    if not has_column("account", "created"):
        op.add_column("account", sa.Column("created", sa.DateTime(), nullable=True))
        op.execute("UPDATE account set account_type='full' WHERE account_type is Null")


def downgrade():
    op.drop_column("account", "created")



"""
Authentication
--------------

User account plugins and authentication.
"""

from . import model
from .interfaces import IClientSkinLayer
from AccessControl import ClassSecurityInfo
from AccessControl.class_init import InitializeClass
from Acquisition import aq_parent
from euphorie.content.user import IUser
from plone import api
from plone.base.utils import safe_text
from plone.keyring.interfaces import IKeyManager
from Products.membrane.interfaces import IMembraneUserAuth
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Products.PluggableAuthService.interfaces.plugins import IAuthenticationPlugin
from Products.PluggableAuthService.interfaces.plugins import IChallengePlugin
from Products.PluggableAuthService.interfaces.plugins import IExtractionPlugin
from Products.PluggableAuthService.interfaces.plugins import IUserEnumerationPlugin
from Products.PluggableAuthService.interfaces.plugins import IUserFactoryPlugin
from Products.PluggableAuthService.plugins.BasePlugin import BasePlugin
from Products.PluggableAuthService.utils import classImplements
from urllib.parse import urlencode
from z3c.saconfig import Session
from zope.component import getUtility
from zope.globalrequest import getRequest
from zope.publisher.interfaces.browser import IBrowserView

import hashlib
import hmac
import logging
import sqlalchemy.exc
import traceback


log = logging.getLogger(__name__)


class NotImplementedError(Exception):
    def __init__(self, message):
        self.message = message


def _get_user(context, login):
    membrane = api.portal.get_tool("membrane_tool")
    user = membrane.getUserObject(login=login)
    if not IUser.providedBy(user):
        return None
    else:
        return user


def generate_token(user):
    """Convenience utility to generate token for the user."""
    manager = getUtility(IKeyManager)
    hasher = hmac.new(manager.secret(), digestmod=hashlib.sha256)
    hasher.update(user.login)
    # Note that unlike authenticate_credentials we access the password
    # directly here. This is fine since here we do not care how the password
    # is stored: even if it is hashed the token will be fine.
    hasher.update(user.password.encode("utf-8"))
    return f"{user.login}-{hasher.hexdigest()}"


def authenticate_cms_token(context, token):
    try:
        (login, hash) = token.split("-")
    except ValueError:
        return None
    user = _get_user(context, login)
    if user is None or generate_token(user) != token:
        return None
    else:
        auth = IMembraneUserAuth(user, None)
        return (auth.getUserId(), auth.getUserName())


def graceful_recovery(default=None, log_args=True):
    """Decorator to safely use SQLAlchemy in PAS plugins. This decorator makes
    sure SQL exceptions are caught and logged.

    Code from Malthe Borch's pas.plugins.sqlalchemy package.
    """

    def decorator(func):
        def wrapper(*args, **kwargs):
            try:
                value = func(*args, **kwargs)
            except sqlalchemy.exc.SQLAlchemyError as e:
                if log_args is False:
                    args = ()
                    kwargs = {}

                formatted_tb = traceback.format_exc()

                try:
                    exc_str = str(e)
                except Exception:
                    exc_str = f"<{e.__class__.__name__} at 0x{id(e):x}>"

                log.critical(
                    "caught SQL-exception: "
                    "%s (in method ``%s``; arguments were %s)\n\n%s"
                    % (
                        exc_str,
                        func.__name__,
                        ", ".join(
                            [repr(arg) for arg in args]
                            + [
                                f"{name}={repr(value)}"
                                for (name, value) in kwargs.items()
                            ]
                        ),
                        formatted_tb,
                    )
                )
                return default
            return value

        return wrapper

    return decorator


manage_addEuphorieAccountPlugin = PageTemplateFile(
    "templates/addPasPlugin", globals(), __name__="manage_addEuphorieAccountPlugin"
)


def addEuphorieAccountPlugin(self, id, title="", REQUEST=None):
    """Add an EuphorieAccountPlugin to a Pluggable Authentication Service."""
    p = EuphorieAccountPlugin(id, title)
    self._setObject(p.getId(), p)

    if REQUEST is not None:
        REQUEST["RESPONSE"].redirect(
            "%s/manage_workspace"
            "?manage_tabs_message=Euphorie+Account+Manager+plugin+added."
            % self.absolute_url()
        )


class EuphorieAccountPlugin(BasePlugin):
    meta_type = "Euphorie account manager"
    security = ClassSecurityInfo()

    def __init__(self, id, title=None):
        self._setId(id)
        self.title = title

    def extractCredentials(self, request):
        """IExtractionPlugin implementation."""
        token = request.getHeader("X-Euphorie-Token")
        if token:
            return {"api-token": token}
        else:
            return {}

    @security.private
    def _authenticate_token(self, credentials):
        """IAuthenticationPlugin implementation."""
        token = credentials.get("api-token")
        if not token:
            return None
        account = authenticate_cms_token(self, token)
        return account

    @security.private
    def _authenticate_login(self, credentials):
        login = credentials.get("login")
        password = credentials.get("password")
        account = authenticate(login, password)
        if account is not None:
            return (str(account.id), account.loginname)
        else:
            return None

    @security.private
    def _get_survey_session(self):
        for parent in self.REQUEST.other["PARENTS"]:
            if isinstance(parent, model.SurveySession):
                return parent
        else:
            return None

    @security.private
    @graceful_recovery(log_args=False)
    def authenticateCredentials(self, credentials):
        if not IClientSkinLayer.providedBy(self.REQUEST):
            return None
        uid_and_login = self._authenticate_login(credentials)
        if uid_and_login is None:
            uid_and_login = self._authenticate_token(credentials)
        if uid_and_login is not None:
            session = self._get_survey_session()
            if session is not None:
                # Verify if current session matches the user. This prevents
                # a cookie hijack attack.
                if str(session.account_id) != uid_and_login[0]:
                    return None
            return uid_and_login
        else:
            return None

    @graceful_recovery()
    def createUser(self, user_id, name):
        """IUserFactoryPlugin implementation."""
        # It only happens with a call from the authomatic plugin that name
        # is empty. In that case, user_id is actually the user's loginname.
        # Force a an explicit search by loginname in that case.
        # Reason: we have users with a loginname that is an integer (!). In such
        # a case, query like `get(user_id)` matches the 'id' column in Account
        # first. If the loginname that is an integer also corresponds to an id
        # in the Account table, we would find the wrong user.
        request = getattr(self, "REQUEST", None) or getRequest()
        if not IClientSkinLayer.providedBy(request):
            # For example: a standard Plone user is being created on the backend.
            return None
        if not name:
            return (
                Session()
                .query(model.Account)
                .filter(model.Account.loginname == user_id)
                .one()
            )
        try:
            user_id = int(user_id)
        except (TypeError, ValueError):
            return None
        return Session().query(model.Account).get(user_id)

    @graceful_recovery()
    def enumerateUsers(
        self,
        id=None,
        login=None,
        exact_match=False,
        sort_by=None,
        max_results=None,
        **kw,
    ):
        """IUserEnumerationPlugin implementation."""
        if not exact_match:
            return []
        if not IClientSkinLayer.providedBy(self.REQUEST):
            return []

        query = Session().query(model.Account)
        if id is not None:
            try:
                query = query.filter(model.Account.id == int(id))
            except ValueError:
                return []
        if login:
            query = query.filter(model.Account.loginname == login)
        account = query.first()
        if account is not None:
            return [{"id": str(account.id), "login": account.loginname}]
        return []

    def updateUser(self, user_id, login_name):
        """Changes the user's username. New method available since Plone 4.3.
        Euphorie doesn't support this.

        :returns: False
        """
        return False

    def updateEveryLoginName(self, quit_on_first_error=True):
        """Update login names of all users to their canonical value.

        This should be done after changing the login_transform
        property of PAS.

        You can set quit_on_first_error to False to report all errors
        before quitting with an error.  This can be useful if you want
        to know how many problems there are, if any.

        :raises: NotImplementedError
        """
        raise NotImplementedError(
            "updateEveryLoginName method is not implemented by Euphorie"
        )

    def challenge(self, request, response):
        """IChallengePlugin implementation."""
        if not IClientSkinLayer.providedBy(request):
            return False

        current_url = request.get("ACTUAL_URL", "")
        query = request.get("QUERY_STRING")
        if query:
            if not query.startswith("?"):
                query = "?" + query
            current_url += query

        context = request.get("PUBLISHED")
        if not context:
            log.error(
                "Refusing to authenticate because no context has been found in %r",  # noqa: E501
                request,
            )
            return False
        if IBrowserView.providedBy(context):
            context = aq_parent(context)

        # In case of a deep link (anything deeper than to the country), open
        # the login form directly.
        deep_link = False
        try:
            webhelpers = api.content.get_view("webhelpers", context, request)
        except Exception:
            pass
        else:
            path = current_url.split(webhelpers.client_url)[-1]
            elems = [item for item in path.split("/") if item]
            if len(elems) > 1:
                deep_link = True

        login_url = "{url}/@@login?{came_from}{fragment}".format(
            url=context.absolute_url(),
            came_from=urlencode(dict(came_from=current_url)),
            fragment="#login" if deep_link else "",
        )
        response.redirect(login_url, lock=True)
        return True


def authenticate(login, password):
    """Try to authenticate a user using the given login and password.

    :param unicode login: login name
    :param unicode password: users password
    :return: :py:class:`Account <euphorie.client.model.Account>` instance

    If the credentials are valid the matching account is returned. For invalid
    credentials None is returned instead.
    """
    if not login or not password:
        return None
    password = safe_text(password)

    login = login.lower()
    accounts = Session().query(model.Account).filter(model.Account.loginname == login)
    for account in accounts:
        if account.verify_password(password):
            return account


classImplements(
    EuphorieAccountPlugin,
    IAuthenticationPlugin,
    IChallengePlugin,
    IExtractionPlugin,
    IUserEnumerationPlugin,
    IUserFactoryPlugin,
)

InitializeClass(EuphorieAccountPlugin)



from euphorie.client import MessageFactory as _
from euphorie.client.interfaces import INotificationCategory
from euphorie.client.interfaces import INTERVAL_DAILY
from euphorie.client.model import Account
from euphorie.client.model import OrganisationMembership
from euphorie.client.model import SurveySession
from euphorie.client.notifications.base import BaseNotification
from euphorie.client.notifications.base import BaseNotificationEmail
from logging import getLogger
from plone import api
from plone.memoize.view import memoize
from sqlalchemy import sql
from z3c.saconfig import Session
from zope.interface import implementer

import datetime


logger = getLogger(__name__)


class Email(BaseNotificationEmail):
    @property
    def main_text(self):
        return _(
            "notification_mail_body__ra_not_modified",
            default="""\
You have not modified your risk assessment for ${reminder_days} days. Please \
remember to keep your risk assessment up to date.
With this link you can access the risk assessment.""",
            msgid_plural="notification_mail_body__ra_not_modified__plural",
            default_plural="""\
You have not modified your risk assessments for ${reminder_days} days. Please \
remember to keep your risk assessments up to date.
With these links you can access the risk assessments.""",
            mapping={
                "reminder_days": self.reminder_days,
            },
            number=len(self.sessions),
        )

    @property
    def translatable_subject(self):
        return _(
            "notification_mail_subject__ra_not_modified",
            default="Reminder: Update of risk assessment (${num} open)",
            mapping={
                "num": len(self.sessions),
            },
        )

    @property
    def reminder_days(self):
        value = api.portal.get_registry_record(
            "euphorie.notification__ra_not_modified__reminder_days", default=365
        )
        return value


@implementer(INotificationCategory)
class Notification(BaseNotification):
    id = "notification__ra_not_modified"
    title = _(
        "notification_title__ra_not_modified",
        default="Notify on orphaned risk assessments.",
    )
    interval = INTERVAL_DAILY
    available = False  # This feature is not yet available

    @property
    @memoize
    def webhelpers(self):
        return api.content.get_view("webhelpers", self.context, self.request)

    @property
    def description(self):
        return _(
            "notification_description__ra_not_modified",
            default="Notify when a risk assessment was not modified for ${days} days.",
            mapping={"days": self.reminder_days},
        )

    @property
    def default(self):
        value = api.portal.get_registry_record(
            "euphorie.notification__ra_not_modified__default", default=False
        )
        return value

    @property
    def reminder_days(self):
        value = api.portal.get_registry_record(
            "euphorie.notification__ra_not_modified__reminder_days", default=365
        )
        return value

    def get_responsible_users(self, session):
        return (
            Session.query(Account)
            .outerjoin(
                OrganisationMembership, OrganisationMembership.member_id == Account.id
            )
            .filter(
                sql.or_(
                    sql.and_(
                        OrganisationMembership.owner_id == session.account_id,
                        OrganisationMembership.member_role.in_(["admin", "manager"]),
                    ),
                    Account.id == session.account_id,
                )
            )
        )

    def notify(self):
        if not self.available:
            return

        today = datetime.date.today()

        base_query = self.webhelpers.get_sessions_query(
            context=self.context,
            include_archived=False,
            filter_by_account=False,
        )
        query = base_query.filter(
            SurveySession.modified
            <= (today - datetime.timedelta(days=self.reminder_days))
        )

        # TODO filter for already sent notifications

        sessions = query.all()

        notifications = {}
        for session in sessions:
            for account in self.get_responsible_users(session):
                if not self.notification_enabled(account):
                    continue
                notifications.setdefault(
                    account.id,
                    {
                        "account": account,
                        "sessions": [],
                    },
                )["sessions"].append(session)

        for notification in notifications.values():
            mail = api.content.get_view(
                name="notification__ra-not-modified__email",
                context={
                    "account": notification["account"],
                    "sessions": notification["sessions"],
                },
                request=self.request,
            )
            mail.send_email()
            logger.info(
                "Sent %s to %s for sessions %s",
                self.id,
                notification["account"].email,
                ", ".join(
                    [
                        session.title or f"{self.title_missing} (id {session.id})"
                        for session in notification["sessions"]
                    ]
                ),
            )



from ftw.upgrade import UpgradeStep


class HideExternalEditorAction(UpgradeStep):
    """Hide external editor action.

    And fix its condition to not fail when the `externalEditorEnabled`
    skin script is not available.
    """

    def __call__(self):
        self.install_upgrade_profile()



from .. import MessageFactory as _
from Acquisition import aq_inner
from Acquisition import aq_parent
from collections import OrderedDict
from io import StringIO
from plone import api
from plone.base.utils import safe_text
from plone.namedfile.interfaces import INamedBlobImage
from plonetheme.nuplone import MessageFactory as NuPloneMessageFactory
from plonetheme.nuplone.utils import checkPermission
from Products.Five import BrowserView
from zope.component import queryUtility
from zope.i18nmessageid.message import Message
from zope.interface import implementer
from zope.interface import Interface
from zope.interface import Invalid
from zope.schema.interfaces import ITitledTokenizedTerm
from zope.schema.interfaces import IVocabularyFactory
from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary

import csv
import re


csv.register_dialect(
    "bilbomatica",
    delimiter=",",
    doublequote=False,
    quoting=csv.QUOTE_ALL,
)

IMAGE_MIN_SIZE = (1000, 430)

TAG = re.compile("<.*?>")
UNWANTED = re.compile("(\r|&#13;|\xad)")
WHITE = re.compile(" +")

REGION_NAMES = {"eu": _("European Union"), "EU": _("European Union")}

CUSTOM_COUNTRY_NAMES = {
    # This are areas whose country codes have not stabilized enough
    # to have arrived in the Unicode consortiums translations,
    # or where we use and old/incorrect country code.
    "CS": _("Kosovo"),
    "ME": _("Montenegro"),
    "RS": _("Republic of Serbia"),
    # Political issue, Macedonia can't be called Macedonia...
    "MK": _("F.Y.R. Macedonia"),
}

# New concept: "Type" or "Flavour" of an OiRA tool. Apart from the "Classic"
# one, we add the type "With measures that are already in place".
link_add_measures = _(
    "no_translate_link_add_measures",
    default="<a class='add-clone' id='add-in-place-measure-button'>${text_add_measures}</a>",  # noqa: E501
    mapping={
        "text_add_measures": _(
            "text_add_measures",
            default="add all measures that have already been implemented",
        )
    },
)

TOOL_TYPES = OrderedDict(
    [
        (
            "classic",
            {
                "use_omega_risks": True,
                "title": _("Classic OiRA Tool with positive and negative statements."),
                "description": "",
                "intro_extra": "",
                "intro_extra_always_present": "",
                "button_add_extra": "",
                "button_add_extra_always_present": "",
                "button_remove_extra": "",
                "button_remove_extra_always_present": "",
                "placeholder_add_extra": "",
                "custom_placeholder_add_extra": "",
                "placeholder_add_extra_always_present": "",
                "intro_questions": "",
                "answer_yes": _("label_yes", default="Yes"),
                "answer_no": _("label_no", default="No"),
                "answer_yes_integrated_ap": _("label_yes_integrated_ap", default="Yes"),
                "answer_no_integrated_ap": _("label_no_integrated_ap", default="No"),
                "answer_na": _("label_not_applicable", default="Not applicable"),
                "custom_intro_extra": "",
                "custom_intro_questions": _(
                    "is_risk_acceptable",
                    default="Is this risk acceptable or under control?",
                ),
                "custom_button_add_extra": "",
            },
        ),
        (
            "existing_measures",
            {
                "use_omega_risks": True,
                "title": _(
                    'OiRA Tool where "Measures that are already in place" can be '
                    "defined."
                ),
                "description": "",
                "intro_extra": _(
                    "select_add_existing_measure",
                    default="Select or add any measures that are <strong>already "
                    "implemented</strong>.",
                ),
                "intro_extra_always_present": _(
                    "select_add_existing_measure",
                    default="Select or add any measures that are <strong>already "
                    "implemented</strong>.",
                ),
                "button_add_extra": _(
                    "button_add_existing_measure",
                    default="Add another measure implemented already",
                ),
                "button_add_extra_always_present": _(
                    "button_add_existing_measure",
                    default="Add another measure implemented already",
                ),
                "button_remove_extra": _(
                    "button_remove_existing_measure",
                    default="Remove this existing measure",
                ),
                "button_remove_extra_always_present": _(
                    "button_remove_existing_measure",
                    default="Remove this existing measure",
                ),
                "placeholder_add_extra": _(
                    "placeholder_new_existing",
                    default="Describe a measure that is already in place but not yet "
                    "listed here.",
                ),
                "custom_placeholder_add_extra": _(
                    "custom_placeholder_new_existing",
                    default="Describe a measure that is already in place.",
                ),
                "placeholder_add_extra_always_present": _(
                    "placeholder_new_existing",
                    default="Describe a measure that is already in place but not yet "
                    "listed here.",
                ),
                "intro_questions": _(
                    "implemented_measures_sufficient",
                    default="Are the measures that are already implemented sufficient?",  # noqa: E501
                ),
                "answer_yes": _(
                    "label_yes_sufficient",
                    default="Yes, the remaining risk is acceptable",
                ),
                "answer_no": _(
                    "label_not_sufficient", default="No, more measures are required"
                ),
                "answer_yes_integrated_ap": _(
                    "label_yes_sufficient",
                    default="Yes, the remaining risk is acceptable",
                ),
                "answer_no_integrated_ap": _(
                    "label_not_sufficient_integrated_ap",
                    default="No, more measures are required (to be added below)",  # noqa: E501
                ),
                "answer_na": _("label_not_applicable", default="Not applicable"),
                "custom_intro_extra": _(
                    "add_existing_measure",
                    default="Add any measures that are <strong>already implemented"
                    "</strong>.",
                ),
                "custom_intro_questions": _(
                    "implemented_measures_sufficient",
                    default="Are the measures that are already implemented sufficient?",  # noqa: E501
                ),
                "custom_button_add_extra": _(
                    "custom_button_add_existing_measure",
                    default="Add an already implemented measure",
                ),
            },
        ),
    ]
)


class IToolTypesInfo(Interface):
    def __call__():
        """Returns info (data-structure) about available tool types."""

    def types_existing_measures():
        """Returns all tool type names that use the "Measures already in place"
        feature."""

    def default_tool_type():
        """Returns the default type of tool, valid for this installation."""


@implementer(IToolTypesInfo)
class ToolTypesInfo:
    def __call__(self):
        return TOOL_TYPES

    @property
    def types_existing_measures(self):
        return ["existing_measures"]

    @property
    def default_tool_type(self):
        # The one defined as first option wins
        for key in TOOL_TYPES:
            return key


def get_tool_type_default():
    tti = queryUtility(IToolTypesInfo)
    if tti is not None:
        return tti.default_tool_type
    return "classic"


@implementer(IVocabularyFactory)
class ToolTypesVocabulary:
    def __call__(self, context):
        tti = queryUtility(IToolTypesInfo)
        terms = []
        if tti is not None:
            for key, val in tti().items():
                terms.append(SimpleTerm(key, title=val["title"]))
        return SimpleVocabulary(terms)


ToolTypesVocabularyFactory = ToolTypesVocabulary()


@implementer(IVocabularyFactory)
class MeasuresTextHandlingVocabulary:
    def __call__(self, context):
        terms = [
            SimpleTerm("full", title=_("Full (show Description and General Approach)")),
            SimpleTerm("simple", title=_("Simple (show only General Approach)")),
        ]
        return SimpleVocabulary(terms)


MeasuresTextHandlingVocabularyFactory = MeasuresTextHandlingVocabulary()


def ensure_image_size(value):
    if INamedBlobImage.providedBy(value):
        img_size = value.getImageSize()
        if img_size < IMAGE_MIN_SIZE:
            msg = f"Image “{value.filename}” is too small. "
            msg += "The minimum size is {} (width) x {} (height) pixels. ".format(
                *IMAGE_MIN_SIZE
            )
            msg += "Your image has a size of {} x {}.".format(*img_size)
            raise Invalid(msg)
    return True


def getSurvey(context):
    from euphorie.content.survey import ISurvey
    from euphorie.content.surveygroup import ISurveyGroup

    obj = context
    while obj and not ISurveyGroup.providedBy(obj):
        if ISurvey.providedBy(obj):
            return obj
        obj = aq_parent(obj)


def StripMarkup(markup):
    """Strip all markup from a HTML fragment."""
    if not markup:
        return ""

    txt = TAG.sub(" ", markup)
    return WHITE.sub(" ", txt).strip()


def StripUnwanted(text):
    """Strip unwanted elements from a text fragment."""
    if not text:
        return ""

    return UNWANTED.sub("", text)


def getTermTitleByValue(field, token):
    try:
        term = field.vocabulary.getTerm(token)
    except LookupError:
        return token

    if ITitledTokenizedTerm.providedBy(term):
        return term.title
    else:
        return term.value


def getTermTitleByToken(field, token):
    try:
        term = field.vocabulary.getTermByToken(str(token))
    except LookupError:
        return token

    if ITitledTokenizedTerm.providedBy(term):
        return term.title
    else:
        return term.value


def getRegionTitle(request, id, default=None):
    names = request.locale.displayNames.territories
    getters = [REGION_NAMES.get, CUSTOM_COUNTRY_NAMES.get, names.get]
    for getter in getters:
        title = getter(id.upper())
        if title is not None:
            if isinstance(title, Message):
                title = api.portal.translate(title)
            return title
    return default if default is not None else id


def summarizeCountries(container, request, current_country=None, permission=None):
    from euphorie.content.country import ICountry

    result = {}
    for country in container.values():
        if not ICountry.providedBy(country):
            continue

        if permission and not checkPermission(country, permission):
            continue

        country_type = getattr(country, "country_type", "country")
        if country_type != "region":
            country_type = "country"
        countries = result.setdefault(country_type, [])
        countries.append(
            {
                "id": country.id,
                "title": getRegionTitle(request, country.id, country.title),
                "url": country.absolute_url(),
                "current": (current_country == country.id),
            }
        )
    for ct in result.values():
        ct.sort(key=lambda c: c["title"])

    return result


class DragDropHelper:
    def sortable_explanation(self):
        return api.portal.translate(
            NuPloneMessageFactory(
                "Change order of items by dragging the handle",
                default="Change order of items by dragging the handle",
            ),
        )


class UserExport(BrowserView):
    def __call__(self):
        ret = "<html><body><h1>User Export</h1>"
        from euphorie.content.countrymanager import ICountryManager
        from euphorie.content.sector import ISector

        for id, country in aq_inner(self.context).objectItems():
            if len(id) != 2:
                continue
            managers = [
                item for item in country.values() if ICountryManager.providedBy(item)
            ]
            sectors = [item for item in country.values() if ISector.providedBy(item)]
            if len(managers) + len(sectors) == 0:
                continue
            ret += f"<h2>{country.title}</h2>"
            ret += "<h3>Country managers</h3><ul>"
            for manager in managers:
                ret += f"<li>{manager.title}, {manager.contact_email}</li>"
            ret += "</ul>"
            ret += "<h3>Sector managers</h3><dl>"
            for sector in sectors:
                ret += "<dt><strong>Sector: {}</strong></dt><dd>{}, {}</dd>".format(
                    sector.title, sector.contact_name, sector.contact_email
                )
            ret += "</dl>"

        ret += "</body></html>"
        return ret


class UserExportCSV(BrowserView):
    def __call__(self):
        from euphorie.content.countrymanager import ICountryManager
        from euphorie.content.sector import ISector

        fieldnames = ["fullname", "email"]
        buffer = StringIO()
        writer = csv.DictWriter(buffer, fieldnames=fieldnames, dialect="bilbomatica")
        writer.writerow({fn: fn for fn in fieldnames})
        for id, country in aq_inner(self.context).objectItems():
            if len(id) != 2:
                continue
            managers = [
                item for item in country.values() if ICountryManager.providedBy(item)
            ]
            sectors = [item for item in country.values() if ISector.providedBy(item)]
            for manager in managers:
                data = dict(
                    fullname=safe_text(manager.title).encode("utf-8"),
                    email=safe_text(manager.contact_email).encode("utf-8"),
                )
                writer.writerow(data)
            for sector in sectors:
                data = dict(
                    fullname=safe_text(sector.contact_name).encode("utf-8"),
                    email=safe_text(sector.contact_email).encode("utf-8"),
                )
                writer.writerow(data)
        csv_data = buffer.getvalue()
        buffer.close()
        response = self.request.RESPONSE
        response.setHeader(
            "Content-Disposition", "attachment; filename=oira_admin_users.csv"
        )
        response.setHeader("Content-Type", "text/csv;charset=utf-8")
        return csv_data


def parse_scaled_answers(contents):
    """Get values and answers from a scaled_answers risk field.

    You pass the contents of the field: a multi-line string.

    In most cases we will get something like this:

        very unsafe
        a bit unsafe
        safe enough
        quite safe
        very safe

    The first answer gets value 1, the second 2, etc.
    But we also support something like this, with the values encoded:

        very safe|5
        quite safe|4
        safe enough|3
        a bit unsafe|2
        very unsafe|1

    """
    if not contents:
        return []

    result = []
    lines = filter(None, (line.strip() for line in contents.splitlines()))

    for idx, line in enumerate(lines, start=1):
        if "|" not in line:
            text, value = line, str(idx)
        else:
            text, value = map(str.strip, line.rpartition("|")[::2])

        result.append(
            {
                "text": text,
                "value": value or str(idx),
            },
        )

    return result



"""
Factory Type Information
------------------------

An extension point for configuring the conditions under which certain portal
types can be created e.g. :obj:`euphorie.content.risk`.
"""

from Acquisition import aq_base
from plone.dexterity.fti import DexterityFTI
from plone.dexterity.interfaces import IDexterityFTI
from zope.component import queryMultiAdapter
from zope.component import queryUtility
from zope.interface import Interface


class IConstructionFilter(Interface):
    """Extra object creation checks.

    This is used by :obj:`ConditionalDexterityFTI` as a named multi-
    adapter, adapting the FTI and the container using the name of the
    portal_type.
    """

    def __init__(fti, container):
        """Adapt on the FTI of the object being created and the target
        container."""

    def allowed():
        """Check if construction is allowed."""


class ConditionalDexterityFTI(DexterityFTI):
    """Special FTI which supports configurable object creation constrains.

    CMF calls to :obj:`isConstructionAllowed` method of the FTI to
    determine of an object of a type can be constructed. In euphorie we
    have some restrictions that only hold for a specific portal type, so
    we need an extension point that can configured per portal type. This
    is done by introducing a new :obj:`IConstructionFilter` named
    adapter.
    """

    def isConstructionAllowed(self, container):
        if not super().isConstructionAllowed(container):
            return False

        filter = queryMultiAdapter((self, container), IConstructionFilter, self.getId())
        if filter is not None:
            return filter.allowed()

        return True


def check_fti_paste_allowed(container, obj):
    """Pasting is only allowed for Dexterity content types which satisty the
    isConstructionAllowed method.

    :param container: [required] container object
    :param obj: [required] the object to be pasted
    :raises: ValueError
    """
    portal_type = getattr(aq_base(obj), "portal_type", None)
    if not portal_type:
        raise ValueError("Can only paste portal content.")
    fti = queryUtility(IDexterityFTI, name=portal_type)
    if fti is None:
        raise ValueError("Can not paste non-dexterity content.")
    if not fti.isConstructionAllowed(container):
        raise ValueError("You can not add the copied content here.")



from functools import cached_property
from z3c.form.browser.select import SelectWidget
from z3c.form.interfaces import IFieldWidget
from z3c.form.widget import FieldWidget
from zope.interface import implementer


class EuphorieChoiceWidget(SelectWidget):
    """Select widget for z3c.form in the Euphorie context: better display
    mode."""

    @cached_property
    def title_value(self):
        if self.name in self.request.form:
            token = self.request.form[self.name]
            vocabulary = self.field.source(self.context)
            try:
                return vocabulary.getTermByToken(token).value
            except LookupError:
                return

        return getattr(self.context.session, self.field.__name__, None)

    def isSelected(self, term):
        return term.title == self.title_value


@implementer(IFieldWidget)
def EuphorieChoiceFieldWidget(field, request):
    """This is identicall to the equivalent in plone.app.z3cform but needs a
    customization to use the EuphorieSelectWidget class."""
    widget = FieldWidget(field, EuphorieChoiceWidget(request))
    return widget



from Acquisition import aq_chain
from Acquisition import aq_inner
from euphorie.content.sector import getSurveys
from euphorie.content.sector import ISector
from plone.tiles import Tile
from plonetheme.nuplone.utils import checkPermission


class SurveyVersions(Tile):
    def update(self):
        for sector in aq_chain(aq_inner(self.context)):
            if ISector.providedBy(sector):
                break
        else:
            sector = aq_inner(self.context)

        self.action_url = "%s/@@version-command" % sector.absolute_url()
        self.surveys = getSurveys(self.context)

    def __call__(self):
        if not checkPermission(self.context, "CMFEditions: Access previous versions"):
            return None

        self.update()
        return self.index()



from datetime import datetime
from euphorie.client.model import Account
from Products.PluggableAuthService.interfaces.events import IUserLoggedInEvent
from zope.component import adapter


@adapter(Account, IUserLoggedInEvent)
def record_last_login(account, event=None):
    account.last_login = datetime.now()



from Acquisition import aq_parent
from euphorie.testing import EuphorieIntegrationTestCase


class ProfileQuestionTests(EuphorieIntegrationTestCase):
    def _create(self, container, *args, **kwargs):
        newid = container.invokeFactory(*args, **kwargs)
        return getattr(container, newid)

    def createProfileQuestion(self):
        country = self.portal.sectors.nl
        sector = self._create(country, "euphorie.sector", "sector")
        surveygroup = self._create(sector, "euphorie.surveygroup", "group")
        survey = self._create(surveygroup, "euphorie.survey", "survey")
        pq = self._create(survey, "euphorie.profilequestion", "profilequestion")
        return pq

    def testNotGloballyAllowed(self):
        self.loginAsPortalOwner()
        types = [fti.id for fti in self.portal.allowedContentTypes()]
        self.assertTrue("euphorie.profilequestion" not in types)

    def testAllowedContentTypes(self):
        self.loginAsPortalOwner()
        pq = self.createProfileQuestion()
        types = [fti.id for fti in pq.allowedContentTypes()]
        self.assertEqual(set(types), {"euphorie.module", "euphorie.risk"})

    def testCanBeCopied(self):
        self.loginAsPortalOwner()
        pq = self.createProfileQuestion()
        self.assertTrue(pq.cb_isCopyable())

    def test_verifyObjectPaste_acceptablePaste(self):
        self.loginAsPortalOwner()
        target = self.createProfileQuestion()
        survey = aq_parent(target)
        source = self._create(survey, "euphorie.module", "other")
        target._verifyObjectPaste(source)

    def test_verifyObjectPaste_block_if_result_too_deep(self):
        self.loginAsPortalOwner()
        target = self.createProfileQuestion()
        survey = aq_parent(target)
        source = self._create(survey, "euphorie.module", "other")
        other = self._create(source, "euphorie.module", "other")
        self._create(other, "euphorie.module", "other")
        self.assertRaises(ValueError, target._verifyObjectPaste, source)



from euphorie.deployment.upgrade.utils import alembic_upgrade
from ftw.upgrade import UpgradeStep


class UpgradeToAlembicHead(UpgradeStep):
    """Create the OrganisationMembership table."""

    def __call__(self):
        alembic_upgrade()



"""
Profile
-------

Various classes and methods to enable the selection screen for profile
questions.
"""

from euphorie.client import model
from euphorie.client.utils import HasText
from euphorie.content.interfaces import ICustomRisksModule
from euphorie.content.interfaces import IQuestionContainer
from euphorie.content.module import IModule
from euphorie.content.profilequestion import IProfileQuestion
from euphorie.content.risk import IFrenchRisk
from euphorie.content.risk import IRisk
from sqlalchemy import sql
from z3c.saconfig import Session


always_present_default = "no"


def AddToTree(
    root,
    node,
    zodb_path=[],
    title=None,
    profile_index=0,
    skip_children=False,
):
    """Add a new node to the session tree.

    :param root: parent node of the new child
    :type root: py:class:`euphorie.client.model.SurveySession` or
          :py:class:`euphorie.client.model.SurveyTreeItem`
    :param node: ZODB object to add to the node.
    :type node: one of the :py:mod:`euphorie.content` content types
    :param zodb_path: list of ids of all parents of the root in the session
           tree
    :param title: title for the generated node. Defaults to the title of the
           ZODB object
    :type title: unicode or None
    :param int profile_index: profile answer index number.
    """
    title = title or node.title

    if title:
        title = title[:500]

    if IQuestionContainer.providedBy(node):
        child = model.Module(title=title, module_id=node.id)
        child.has_description = HasText(node.description)
        if IModule.providedBy(node):
            child.solution_direction = HasText(node.solution_direction)
            # All optional modules default to "skip". The user needs to
            # actively decide that the module is relevant for them.
            if node.optional:
                child.skip_children = True
                child.has_description = True
            else:
                child.postponed = False
    elif IRisk.providedBy(node):
        priority = getattr(node, "default_priority", None)
        if priority == "none":
            priority = None

        if IFrenchRisk.providedBy(node):
            effect = node.default_severity
        else:
            effect = node.default_effect

        child = model.Risk(
            title=title,
            risk_id=node.id,
            risk_type=node.type,
            skip_evaluation=(node.evaluation_method == "fixed"),
            probability=node.default_probability,
            frequency=node.default_frequency,
            effect=effect,
            priority=priority,
        )
        child.skip_children = False
        child.postponed = False
        child.has_description = HasText(node.description)
        if node.type in ["top5", "policy"]:
            child.priority = "high"
        if node.risk_always_present and always_present_default:
            child.identification = always_present_default
    else:
        return None  # Should never happen

    zodb_path = zodb_path + [node.id]
    child.zodb_path = "/".join(zodb_path)
    child.profile_index = profile_index
    root.addChild(child)

    if IQuestionContainer.providedBy(node) and not skip_children:
        for grandchild in node.values():
            AddToTree(child, grandchild, zodb_path, None, profile_index)
    return child


def get_custom_risks(session):
    if session is None:
        return []
    query = (
        Session.query(model.Risk)
        .filter(
            sql.and_(
                model.Risk.is_custom_risk == True,  # noqa: E712
                model.Risk.path.startswith(model.Module.path),
                model.Risk.session_id == session.id,
            )
        )
        .order_by(model.Risk.id)
    )
    return query.all()


def BuildSurveyTree(survey, profile, dbsession, old_session=None):
    """(Re)build the survey SQL tree. The existing tree for the session is
    deleted before a new tree is created.

    :param survey: survey to build tree for
    :type survey: :py:class:`euphorie.content.survey.Survey`
    :param profile: desired profile to be used for the tree
    :type profile: dictionary
    :param dbsession: session to build tree in. Defaults to currently active
          session.
    :type dbsession: :py:class:`euphorie.client.model.SurveySession`
    """
    dbsession.reset()

    for child in survey.values():
        if ICustomRisksModule.providedBy(child):
            AddToTree(dbsession, child)
            # Now copy over the custom risks
            risks = get_custom_risks(old_session)
            if risks:
                # find the module that holds the custom risks
                modules = (
                    Session.query(model.Module)
                    .filter(
                        sql.and_(
                            model.Module.session_id == dbsession.id,
                            model.Module.module_id == child.id,
                        )
                    )
                    .all()
                )
                # there should only ever be 1 result
                if modules:
                    for risk in risks:
                        modules[0].addChild(risk)
        elif IProfileQuestion.providedBy(child):
            # Safeguard against double adding of profile questions
            existing = [getattr(item, "module_id") for item in dbsession.children()]
            if child.id in existing:
                continue
            p = profile.get(child.id)
            if not p:
                continue

            if isinstance(p, list):
                profile_question = AddToTree(
                    dbsession,
                    child,
                    title=child.title,
                    profile_index=-1,
                    skip_children=True,
                )
                for index, title in enumerate(p):
                    AddToTree(profile_question, child, title=title, profile_index=index)
            # If we get a bool, it will be True, because of `if not p` above
            # Simply add the profile to the tree, don't care about locations
            elif isinstance(p, bool):
                AddToTree(dbsession, child, title=child.title)
        else:
            AddToTree(dbsession, child)


def extractProfile(survey, survey_session):
    """Determine the current profile for given current survey session.

    :param survey: current survey
    :type survey: :py:class:`euphorie.content.survey.Survey`
    :param survey_session: current survey session
    :type survey_session: :py:class:`euphorie.client.model.SurveySession`
    :rtype: dictionary with profile answers

    The profile is returned as a dictionary. The id of the profile questions
    are used as keys. For optional profile questions the value is a boolean.
    For repetable profile questions the value is a list of titles as provided
    by the user. This format is compatible with
    :py:meth:`Profile.getDesiredProfile`.
    """
    questions = [
        {"id": child.id, "use_location_question": child.use_location_question}
        for child in survey.ProfileQuestions()
    ]
    if not questions:
        return {}

    q_ids = [q["id"] for q in questions]
    session_modules = {}
    query = (
        Session.query(model.SurveyTreeItem.zodb_path, model.SurveyTreeItem.title)
        .filter(model.SurveyTreeItem.type == "module")
        .filter(model.SurveyTreeItem.session == survey_session)
        .filter(model.SurveyTreeItem.profile_index >= 0)
        .filter(model.SurveyTreeItem.zodb_path.in_(q_ids))
        .order_by(model.SurveyTreeItem.profile_index)
    )
    for row in query:
        session_modules.setdefault(row.zodb_path, []).append(row)

    profile = {}
    for question in questions:
        nodes = session_modules.get(question["id"], [])
        if not question["use_location_question"]:
            profile[question["id"]] = bool(nodes)
        else:
            profile[question["id"]] = [node.title for node in nodes]

    return profile



from datetime import date
from euphorie.client import model
from euphorie.client.interfaces import IClientSkinLayer
from euphorie.content.risk import IRisk
from euphorie.content.solution import ISolution
from euphorie.deployment.upgrade.utils import alembic_upgrade_to
from json import loads
from plone import api
from plone.dexterity.interfaces import IDexterityContainer
from sqlalchemy import and_
from transaction import commit
from z3c.saconfig import Session
from zope.interface import alsoProvides

import logging


log = logging.getLogger(__name__)


def alembic_upgrade(context):
    alembic_upgrade_to("27")


def walk(node):
    for idx, sub_node in node.ZopeFind(node, search_sub=0):
        if ISolution.providedBy(sub_node):
            yield sub_node
        if IDexterityContainer.providedBy(sub_node):
            yield from walk(sub_node)


def unifiy_fields(walker):
    count = 0
    for solution in walker:
        count += 1
        if not getattr(solution, "action", None):
            # Only take action if the "action" field has not been set
            action = solution.action_plan.strip()
            prevention_plan = solution.prevention_plan
            prevention_plan = prevention_plan and prevention_plan.strip() or ""
            if prevention_plan:
                action = f"{action}\n{prevention_plan}"
            solution.action = action
            if count % 100 == 0:
                log.info("Handled %d items" % count)
            if count % 1000 == 0:
                log.info("Intermediate commit")
                commit()
        else:
            if count % 100 == 0:
                log.info(
                    "Skipped %d items, since they have already been handled." % count
                )
            if count % 1000 == 0:
                log.info("Intermediate commit")
                commit()
    log.info("Finished. Updated %d solutions" % count)


def unify_action_fields_in_solution(context):
    site = api.portal.get()
    section = "sectors"
    walker = walk(getattr(site, section))
    log.info(f'Iterating over section "{section}"')
    unifiy_fields(walker)


def unify_action_fields_in_solution_client(context):
    site = api.portal.get()
    section = "client"
    walker = walk(getattr(site, section))
    log.info(f'Iterating over section "{section}"')
    unifiy_fields(walker)


def get_pre_defined_measures(solutions, country):
    """Iterate over the Solution items on this risk."""
    measures = {}

    for item in solutions:
        description = item.description and item.description.strip() or ""
        prevention_plan = item.prevention_plan and item.prevention_plan.strip() or ""
        measure = description
        if country in ("it",):
            if prevention_plan:
                measure = f"{measure}: {prevention_plan}"
        measures[measure] = item.id

    return measures


def migrate_actgion_plans(context):
    # Work in chunks of 25
    max_tools = 25
    site = api.portal.get()
    client = getattr(site, "client")
    today = date.today()
    request = context.REQUEST.clone()
    alsoProvides(request, IClientSkinLayer)
    # path = "eu/maritime-transport/maritime-transport-1"
    tool_paths = Session.query(model.SurveySession.zodb_path).distinct()
    # .filter(model.SurveySession.zodb_path==path).distinct()
    tool_count = 0
    skip_count = 0
    for result in tool_paths:
        tool_path = str(result[0])
        try:
            tool = client.restrictedTraverse(tool_path)
        except KeyError:
            log.warning(f"No tool in client found for {tool_path}")
            continue
        country = tool_path.split("/")[0]
        sessions = (
            Session.query(model.SurveySession)
            .filter(
                and_(
                    model.Account.id == model.SurveySession.account_id,
                    model.Account.account_type != "guest",
                    model.SurveySession.zodb_path == tool_path,
                    model.SurveySession.migrated == None,  # noqa: E711
                )
            )
            .order_by(model.SurveySession.zodb_path)
        )
        if not sessions.count():
            skip_count += 1
            if skip_count % 5 == 0:
                log.info("Skipped the first %d, already handled" % skip_count)
            continue
        log.info(f"\n\nHandle tool {tool_path}")
        risks_by_path = {}
        solutions_by_path = {}
        measures_by_path = {}
        for session in sessions:
            log.info(f"Session {session.id}")
            risks = (
                Session.query(model.Risk)
                .filter(model.Risk.session_id == session.id)
                .order_by(model.Risk.path)
            )
            for risk in risks:
                risk_path = str(risk.zodb_path)
                is_custom = "custom" in risk_path
                if is_custom:
                    zodb_risk = None
                elif risk_path not in risks_by_path:
                    try:
                        zodb_risk = tool.restrictedTraverse(risk_path)
                    except Exception:
                        log.warning(f"Risk {risk_path} not found in tool {tool_path}")
                        continue
                    else:
                        if not IRisk.providedBy(zodb_risk):
                            zodb_risk = None
                            solutions_by_path[risk_path] = []
                            measures_by_path[risk_path] = []
                        else:
                            solutions = zodb_risk._solutions
                            solutions_by_path[risk_path] = solutions
                            measures_by_path[risk_path] = get_pre_defined_measures(
                                solutions, country
                            )
                        risks_by_path[risk_path] = zodb_risk

                else:
                    zodb_risk = risks_by_path[risk_path]

                # convert ActionPlan items that are clearly based on solutions
                # to "standard" type
                if not is_custom:
                    for ap in risk.custom_measures:
                        for solution in solutions_by_path[risk_path]:
                            if solution.action_plan == (ap.action_plan or "").strip():
                                if (
                                    getattr(solution, "prevention_plan", "") or ""
                                ).strip() == (ap.prevention_plan or "") and (
                                    getattr(solution, "requirements", "") or ""
                                ).strip() == (
                                    ap.requirements or ""
                                ):
                                    ap.plan_type = "measure_standard"
                                    ap.solution_id = solution.id
                # Convert the measures-in-place to their respective ActionPlan items
                try:
                    saved_existing_measures = (
                        risk.existing_measures and loads(risk.existing_measures) or []
                    )
                except ValueError:
                    saved_existing_measures = []
                # Backwards compat. We used to save dicts before we
                # switched to list of tuples.
                if isinstance(saved_existing_measures, dict):
                    saved_existing_measures = [
                        (k, v) for (k, v) in saved_existing_measures.items()
                    ]
                new_action_plans = []
                already_converted_measure_ids = [
                    x.solution_id for x in risk.in_place_standard_measures
                ]
                already_converted_custom_texts = [
                    x.action for x in risk.in_place_custom_measures
                ]
                if saved_existing_measures:
                    custom = []
                    while saved_existing_measures:
                        text, active = saved_existing_measures.pop()
                        if not active:
                            continue
                        if not is_custom and text in measures_by_path[risk_path]:
                            solution_id = measures_by_path[risk_path][text]
                            if solution_id not in already_converted_measure_ids:
                                new_action_plans.append(
                                    model.ActionPlan(
                                        action=text,
                                        plan_type="in_place_standard",
                                        solution_id=solution_id,
                                    )
                                )
                        else:
                            if text not in already_converted_custom_texts:
                                custom.append(text)
                    for text in custom:
                        new_action_plans.append(
                            model.ActionPlan(action=text, plan_type="in_place_custom")
                        )
                if new_action_plans:
                    risk.action_plans.extend(new_action_plans)
            session.migrated = today
        tool_count += 1
        if max_tools > 0 and tool_count >= max_tools:
            log.info("Broke off after %d tools" % tool_count)
            return



"""Allow consultant deletion.

Revision ID: 20230515143459
Revises: 20230504061955
Create Date: 2023-05-15 14:36:08.134660
"""

from alembic import op

import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "20230515143459"
down_revision = "20230504061955"
branch_labels = None
depends_on = None


def upgrade():
    try:
        op.alter_column(
            "consultancy", "account_id", existing_type=sa.INTEGER(), nullable=True
        )
        op.drop_constraint(
            "consultancy_account_id_fkey", "consultancy", type_="foreignkey"
        )
        op.create_foreign_key(
            None,
            "consultancy",
            "account",
            ["account_id"],
            ["id"],
            onupdate="CASCADE",
            ondelete="SET NULL",
        )
    except Exception:
        # SQLite
        pass


def downgrade():
    op.drop_constraint(None, "consultancy", type_="foreignkey")
    op.create_foreign_key(
        "consultancy_account_id_fkey",
        "consultancy",
        "account",
        ["account_id"],
        ["id"],
        onupdate="CASCADE",
        ondelete="CASCADE",
    )
    op.alter_column(
        "consultancy", "account_id", existing_type=sa.INTEGER(), nullable=False
    )



from contextlib import contextmanager
from euphorie.client import model
from euphorie.client.interfaces import IClientSkinLayer
from euphorie.client.utils import getRequest
from euphorie.client.utils import setRequest
from plone import api
from plone.app.testing import FunctionalTesting
from plone.app.testing import IntegrationTesting
from plone.app.testing import login
from plone.app.testing import logout
from plone.app.testing import PLONE_FIXTURE
from plone.app.testing import PloneSandboxLayer
from plone.app.testing import quickInstallProduct
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import TEST_USER_PASSWORD
from plone.testing import zope
from plone.testing.zope import Browser
from sqlalchemy import event
from transaction import commit
from unittest import mock
from unittest import TestCase
from z3c.saconfig import Session
from Zope2.Startup.datatypes import default_zpublisher_encoding
from zope.interface import alsoProvides
from zope.sqlalchemy.datamanager import NO_SAVEPOINT_SUPPORT

import euphorie.deployment


NO_SAVEPOINT_SUPPORT.remove("sqlite")


class EuphorieFixture(PloneSandboxLayer):
    saconfig_filename = "configure.zcml"

    defaultBases = (PLONE_FIXTURE,)

    def setUpZope(self, app, configurationContext):
        # Load any other ZCML that is required for your tests.
        # The z3c.autoinclude feature is disabled in the Plone fixture base
        # layer.
        import euphorie.client.tests
        import euphorie.deployment.tests

        self.loadZCML("configure.zcml", package=euphorie.deployment)
        self.loadZCML("overrides.zcml", package=euphorie.deployment)
        self.loadZCML("configure.zcml", package=euphorie.client.tests)
        self.loadZCML("configure.zcml", package=euphorie.deployment.tests)
        zope.installProduct(app, "Products.membrane")

        self.loadZCML(self.saconfig_filename, package=euphorie.client.tests)
        engine = Session.bind

        @event.listens_for(engine, "connect")
        def do_connect(dbapi_connection, connection_record):
            # disable pysqlite's emitting of the BEGIN statement entirely.
            # also stops it from emitting COMMIT before any DDL.
            dbapi_connection.isolation_level = None

        @event.listens_for(engine, "begin")
        def do_begin(conn):
            # emit our own BEGIN
            conn.execute("BEGIN")

        # Start fresh
        self.testTearDown()

        default_zpublisher_encoding("utf-8")

    def setUpPloneSite(self, portal):
        pw = api.portal.get_tool("portal_workflow")
        pw.setDefaultChain("simple_publication_workflow")

        quickInstallProduct(portal, "plonetheme.nuplone")
        quickInstallProduct(portal, "euphorie.client")
        quickInstallProduct(portal, "euphorie.content")

        all_countries = euphorie.deployment.setuphandlers.COUNTRIES.copy()
        euphorie.deployment.setuphandlers.COUNTRIES = {
            key: all_countries[key]
            for key in all_countries
            if key
            in (
                "de",
                "nl",
            )
        }

        with mock.patch(
            "plone.i18n.utility.LanguageUtility.listSupportedLanguages",
            return_value=[
                ("de", "German"),
                ("nl", "Dutch"),
            ],
        ):
            quickInstallProduct(portal, "euphorie.deployment.tests")

        session = portal.acl_users.get("session")
        # In tests we don't have secure connections
        session.secure = False

    def testSetUp(self):
        """XXX testSetUp and testTearDown should not be necessary, but it seems
        SQL data is not correctly cleared at the end of a test method run, even
        if testTearDown does an explicit transaction.abort()"""
        model.metadata.create_all(Session.bind, checkfirst=True)

    def testTearDown(self):
        Session.remove()
        model.metadata.drop_all(Session.bind)


class EuphorieRobotFixture(EuphorieFixture):
    saconfig_filename = "robot.zcml"


EUPHORIE_FIXTURE = EuphorieFixture()
EUPHORIE_ROBOT_FIXTURE = EuphorieRobotFixture()

EUPHORIE_INTEGRATION_TESTING = IntegrationTesting(
    bases=(EUPHORIE_FIXTURE,),
    name="EuphorieFixture:Integration",
)


EUPHORIE_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(EUPHORIE_FIXTURE,),
    name="EuphorieFixture:Functional",
)


class EuphorieIntegrationTestCase(TestCase):
    layer = EUPHORIE_INTEGRATION_TESTING
    request_layer = IClientSkinLayer

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]

    def loginAsPortalOwner(self):
        return login(self.app, "admin")

    def login(self, username):
        return login(self.portal, username)

    def logout(self):
        return logout()

    @contextmanager
    def _get_view(self, name, obj, survey_session=None, client=None):
        """Get's a view with a proper fresh request.

        If survey_session is set the SessionManager will be configured
        """
        old_request = getRequest()
        request = self.get_client_request(client=client)
        if survey_session is not None:
            request.other["euphorie.session"] = survey_session
        try:
            setRequest(request)
            yield api.content.get_view(name, obj, request)
        finally:
            setRequest(old_request)

    def get_client_request(self, client=None):
        request = self.request.clone()
        request.client = client
        alsoProvides(request, self.request_layer)
        return request


class EuphorieFunctionalTestCase(EuphorieIntegrationTestCase):
    layer = EUPHORIE_FUNCTIONAL_TESTING

    _default_credentials = {
        "username": TEST_USER_NAME,
        "password": TEST_USER_PASSWORD,
    }

    def get_browser(self, logged_in=False, credentials={}):
        """Return a browser, potentially a logged in one.

        The default credentials are the admin ones
        """
        commit()
        browser = Browser(self.app)
        if logged_in or credentials:
            username = credentials.get(
                "username", self._default_credentials["username"]
            )
            password = credentials.get(
                "password", self._default_credentials["password"]
            )
            browser.open("%s/@@login" % self.portal.absolute_url())
            browser.getControl(name="__ac_name").value = username
            browser.getControl(name="__ac_password").value = password
            browser.getForm(id="loginForm").submit()
        return browser



"""
Export Survey
-------------

Browser view for exporting a complete survey as XML.

View name: @@export
"""

from Acquisition import aq_inner
from Acquisition import aq_parent
from bs4 import BeautifulSoup
from euphorie.client.utils import HasText
from euphorie.content import MessageFactory as _
from euphorie.content.behaviors.toolcategory import IToolCategory
from euphorie.content.browser.upload import COMMA_REPLACEMENT
from euphorie.content.browser.upload import NSMAP
from euphorie.content.browser.upload import ProfileQuestionLocationFields
from euphorie.content.module import IModule
from euphorie.content.profilequestion import IProfileQuestion
from euphorie.content.risk import IKinneyEvaluation
from euphorie.content.risk import IRisk
from euphorie.content.solution import ISolution
from euphorie.content.survey import get_tool_type
from euphorie.content.training_question import ITrainingQuestion
from euphorie.content.utils import StripMarkup
from euphorie.content.utils import StripUnwanted
from io import BytesIO
from lxml import etree
from lxml import html
from plone import api
from plone.autoform.form import AutoExtensibleForm
from plone.base.utils import safe_bytes
from plonetheme.nuplone.z3cform.directives import depends
from z3c.form import button
from z3c.form import form
from zipfile import ZipFile
from zope import schema
from zope.interface import Interface

import re


all_breaks = re.compile("(\n|\r)+")

try:
    from base64 import encodebytes
except ImportError:
    # PY27
    from base64 import encodestring as encodebytes


def getToken(field, value, default=None):
    """Looks up a term in the vocabulary associated with a field and returns
    its token, or raises a LookupError.

    :rtype: str
    :raises: LookupError
    """
    try:
        return field.vocabulary.getTerm(value).token
    except LookupError:
        return default


class IExportSurveySchema(Interface):
    """Fields used for configuring the export."""

    include_images = schema.Bool(
        title=_("label_survey_export_include_images", default="Include images"),
        description=_(
            "description_survey_export_include_images",
            default="All images contained in the OiRA tool can be exported along "
            "with the textual content. When the OiRA tool is then imported again, "
            "all images will be present. Please note that including images may "
            "result in a considerably larger file size.",
        ),
        required=False,
        default=False,
    )
    include_intro_text = schema.Bool(
        title=_(
            "label_survey_export_include_intro_text",
            default="Include introduction text",
        ),
        description=_(
            "description_survey_export_include_intro_text",
            default="The introduction text describes the general purpose of the OiRA "
            "tool. It usually gives the end users an idea what to expect in the tool "
            "and whether it will be relevant for them. The default is to include "
            "this text.",
        ),
        required=False,
        default=True,
    )
    include_module_solution_texts = schema.Bool(
        title=_(
            "label_survey_export_include_module_solution_texts",
            default="Include “Solution” texts on modules",
        ),
        description=_(
            "description_survey_export_include_module_solution_texts",
            "The “Solution” description” on modules is a deprecated kind of text "
            "that is no longer being used. The default is to skip this.",
        ),
        required=False,
        default=False,
    )
    include_module_description_texts = schema.Bool(
        title=_(
            "label_survey_export_include_module_description_texts",
            default="Include “Description” texts on modules",
        ),
        description=_(
            "description_survey_export_include_module_description_texts",
            default="Every module comes with an introductive description. The default "
            "is to include this text.",
        ),
        required=False,
        default=True,
    )
    include_risk_legal_texts = schema.Bool(
        title=_(
            "label_survey_export_include_risk_legal_texts",
            default="Include “Legal and policy references” texts on risks",
        ),
        description=_(
            "description_survey_export_include_risk_legal_texts",
            default="If the OiRA tool is exported because it should be made available "
            "for a different country, the legal texts might not apply there. Therefore "
            "the default is to skip these texts.",
        ),
        required=False,
        default=True,
    )
    include_measures = schema.Bool(
        title=_("label_survey_export_include_measures", default="Include measures"),
        description=_(
            "description_survey_export_include_measures",
            default="For each risk, a number of common measures – sometimes also "
            "referred to as “Solutions” – can be defined. While the default setting "
            "is to include them in the export, they may be excluded if required, "
            "for example to accommodate for a limited translation budget.",
        ),
        required=False,
        default=True,
    )

    depends("export_as_plain_text", "include_images", "off")
    export_as_plain_text = schema.Bool(
        title=_(
            "label_export_as_plain_text",
            default="Export additionally as plain text file",
        ),
        description=_(
            "description_export_as_plain_text",
            default="The plain text file will contain all translatable text without "
            "formatting. This file can be used to determine a word count for "
            "estimating the cost of translation.",
        ),
        required=False,
        default=False,
    )
    is_etranslate_compatible = schema.Bool(
        title=_(
            "label_export_etranslate_compatible",
            default="Export in eTranslate compatible XML",
        ),
        description=_(
            "description_is_etranslate_compatible",
            default="The default export escapes HTML, which is not supported"
            "by eTranslate",
        ),
        required=False,
        default=False,
    )


class ExportSurvey(AutoExtensibleForm, form.Form):
    """The upload view for a :obj:`euphorie.content.sector`

    View name: @@upload
    """

    schema = IExportSurveySchema
    ignoreContext = True
    form_name = _("Export OiRA Tool")
    # defaults
    include_images = False
    include_intro_text = True
    include_module_solution_texts = False
    include_module_description_texts = True
    include_risk_legal_texts = False
    include_measures = True
    export_as_plain_text = False
    is_etranslate_compatible = False

    def _add_string_or_html(self, node, text, element):
        stripped_text = StripUnwanted(text)
        if self.is_etranslate_compatible:
            html_fragment = html.fragment_fromstring(stripped_text, element)
            node.append(html_fragment)
        else:
            etree.SubElement(node, element).text = stripped_text
        return node

    @button.buttonAndHandler(_("Export"))
    def handleExport(self, action):
        (data, errors) = self.extractData()
        self.include_images = data.get("include_images")
        self.include_intro_text = data.get("include_intro_text")
        self.include_module_solution_texts = data.get("include_module_solution_texts")
        self.include_module_description_texts = data.get(
            "include_module_description_texts"
        )
        self.include_risk_legal_texts = data.get("include_risk_legal_texts")
        self.include_measures = data.get("include_measures")
        self.export_as_plain_text = data.get("export_as_plain_text")
        self.is_etranslate_compatible = data.get("is_etranslate_compatible")
        rendered_output = self.render_output()
        self.request.response.write(rendered_output)

    @button.buttonAndHandler(_("button_cancel", default="Cancel"), name="cancel")
    def handleCancel(self, action):
        state = api.content.get_view(
            "plone_context_state", aq_inner(self.context), self.request
        )
        self.request.response.redirect(state.view_url())

    def render_output(self):
        output = etree.Element("sector", nsmap=NSMAP)
        self.exportSurvey(output, self.context)
        response = self.request.response
        rendered_xml = etree.tostring(
            output, pretty_print=True, xml_declaration=True, encoding="utf-8"
        )
        context_id = aq_parent(aq_inner(self.context)).id
        if self.export_as_plain_text and not self.include_images:
            outer_soup = BeautifulSoup(rendered_xml, "lxml")
            # Some tags of the survey contain only functional information, such
            # as true/false or calculated/evaluated
            # Strip them for the word counting
            functional_tags = (
                "classification-code",
                "evaluation-method",
                "evaluation-algorithm",
                "evaluation-optional",
                "integrated_action_plan",
                "language",
                "measures_text_handling",
                "show-not-applicable",
                "tool_type",
            )
            for tag in functional_tags:
                entities = outer_soup.findAll(tag)
                [entity.extract() for entity in entities]
            # The extracted texts contain HTML entities such as &lt;p&gt;
            # We need to convert them to real tags to be able to strip them too for
            # word counting
            inner_soup = BeautifulSoup(outer_soup.get_text(), "html.parser")
            text = inner_soup.get_text()
            text = all_breaks.sub(" ", text)
            zip_output = BytesIO()
            with ZipFile(zip_output, mode="w") as zip:
                zip.writestr(f"{context_id}_word_count.txt", text)
                zip.writestr(f"{context_id}.xml", rendered_xml)
            self.request.response.setHeader("Content-Type", "application/zip")
            self.request.response.setHeader(
                "Content-Disposition", f'attachment; filename="{context_id}.zip"'
            )
            return zip_output.getvalue()
        else:
            response.setHeader(
                "Content-Disposition", f'attachment; filename="{context_id}.xml"'
            )
            response.setHeader("Content-Type", "text/xml")
            return rendered_xml

    def render(self):
        if self.request.response.headers.get("content-disposition", "").startswith(
            "attachment"
        ):
            # We just generated an attachment, no need to render the form
            return ""
        return super().render()

    def exportImage(self, parent, image, caption=None, tagname="image"):
        """:returns: base64 encoded image."""
        if not self.include_images:
            return
        node = etree.SubElement(parent, tagname)
        if image.contentType:
            node.attrib["content-type"] = image.contentType
        if image.filename:
            node.attrib["filename"] = image.filename
        if caption:
            node.attrib["caption"] = caption
        node.text = encodebytes(safe_bytes(image.data))
        return node

    def exportSurvey(self, parent, survey):
        """Export a survey given a parent and the survey itself.

        :returns: An XML node with the details of an :obj:`euphorie.content.survey`.
        """
        node = etree.SubElement(parent, "survey")
        if getattr(survey, "external_id", None):
            node.attrib["external-id"] = survey.external_id
        etree.SubElement(node, "title").text = aq_parent(survey).title
        if self.include_intro_text and StripMarkup(survey.introduction):
            node = self._add_string_or_html(node, survey.introduction, "introduction")
        if survey.classification_code:
            etree.SubElement(node, "classification-code").text = (
                survey.classification_code
            )
        etree.SubElement(node, "language").text = survey.language
        report_completion_threshold = getattr(
            survey, "report_completion_threshold", False
        )
        if report_completion_threshold:
            etree.SubElement(
                node,
                "report_completion_threshold",
                attrib={"value": str(report_completion_threshold)},
            )
        enable_web_training = getattr(survey, "enable_web_training", False)
        if enable_web_training:
            etree.SubElement(node, "enable_web_training", attrib={"value": "true"})

        enable_test_questions = getattr(survey, "enable_test_questions", False)
        if enable_test_questions:
            etree.SubElement(node, "enable_test_questions", attrib={"value": "true"})

        enable_email_reminder = getattr(survey, "enable_email_reminder", False)
        if enable_email_reminder:
            etree.SubElement(node, "enable_email_reminder", attrib={"value": "true"})
        num_training_questions = getattr(survey, "num_training_questions", False)
        if num_training_questions:
            etree.SubElement(
                node,
                "num_training_questions",
                attrib={"value": str(num_training_questions)},
            )
        if self.is_etranslate_compatible:
            etree.SubElement(node, "tool_type", attrib={"value": get_tool_type(survey)})
            etree.SubElement(
                node,
                "measures_text_handling",
                attrib={"value": getattr(survey, "measures_text_handling", "full")},
            )
            etree.SubElement(
                node,
                "integrated_action_plan",
                attrib={
                    "value": (
                        "true"
                        if getattr(survey, "integrated_action_plan", False)
                        else "false"
                    )
                },
            )
            etree.SubElement(
                node,
                "evaluation-algorithm",
                attrib={"value": aq_parent(survey).evaluation_algorithm},
            )
            etree.SubElement(
                node,
                "evaluation-optional",
                attrib={"value": "true" if survey.evaluation_optional else "false"},
            )
        else:
            etree.SubElement(node, "tool_type").text = get_tool_type(survey)
            etree.SubElement(node, "measures_text_handling").text = getattr(
                survey, "measures_text_handling", "full"
            )
            etree.SubElement(node, "integrated_action_plan").text = (
                "true" if getattr(survey, "integrated_action_plan", False) else "false"
            )
            etree.SubElement(node, "evaluation-algorithm").text = aq_parent(
                survey
            ).evaluation_algorithm
            etree.SubElement(node, "evaluation-optional").text = (
                "true" if survey.evaluation_optional else "false"
            )
        if IToolCategory.providedBy(survey):
            tool_category = IToolCategory(survey).tool_category or []
            etree.SubElement(node, "tool-category").text = ", ".join(
                [x.replace(",", COMMA_REPLACEMENT) for x in tool_category]
            )

        if getattr(survey, "external_site_logo", None):
            self.exportImage(
                node, survey.external_site_logo, tagname="external_site_logo"
            )
        if getattr(survey, "image", None):
            self.exportImage(node, survey.image)

        for child in survey.values():
            if IProfileQuestion.providedBy(child):
                self.exportProfileQuestion(node, child)
            if IModule.providedBy(child):
                self.exportModule(node, child)
            if ITrainingQuestion.providedBy(child):
                self.exportTrainingQuestion(node, child)
        return node

    def exportProfileQuestion(self, parent, profile):
        """:returns: An XML node with the details of an
        :obj:`euphorie.content.profilequestion`.
        """
        node = etree.SubElement(parent, "profile-question")
        if getattr(profile, "external_id", None):
            node.attrib["external-id"] = profile.external_id
        etree.SubElement(node, "title").text = profile.title
        # Use title if question is not available (Euphorie < 2.0rc2 data)
        if HasText(profile.question):
            node = self._add_string_or_html(node, profile.question, "question")
        else:
            etree.SubElement(node, "question").text = profile.title
        if HasText(profile.description):
            node = self._add_string_or_html(node, profile.description, "description")
        for fname in ProfileQuestionLocationFields:
            value = getattr(profile, fname, None)
            if value:
                etree.SubElement(node, fname.replace("_", "-")).text = value

        etree.SubElement(node, "use-location-question").text = (
            "true" if getattr(profile, "use_location_question", True) else "false"
        )

        for child in profile.values():
            if IModule.providedBy(child):
                self.exportModule(node, child)
            elif IRisk.providedBy(child):
                self.exportRisk(node, child)
        return node

    def exportModule(self, parent, module):
        """:returns: An XML node with the details of an
        :obj:`euphorie.content.module`."""

        module_kwargs = {"optional": "true" if module.optional else "false"}
        if getattr(module, "hide_from_training", False):
            module_kwargs["hide_from_training"] = "true"
        node = etree.SubElement(parent, "module", **module_kwargs)

        if getattr(module, "external_id", None):
            node.attrib["external-id"] = module.external_id
        etree.SubElement(node, "title").text = module.title
        if self.include_module_description_texts and HasText(module.description):
            node = self._add_string_or_html(node, module.description, "description")
        if module.optional:
            etree.SubElement(node, "question").text = module.question
        if self.include_module_solution_texts and StripMarkup(
            module.solution_direction
        ):
            etree.SubElement(node, "solution-direction").text = StripUnwanted(
                module.solution_direction
            )
        if module.image is not None:
            self.exportImage(node, module.image, module.caption)

        for child in module.values():
            if IModule.providedBy(child):
                self.exportModule(node, child)
            elif IRisk.providedBy(child):
                self.exportRisk(node, child)
        return node

    def exportRisk(self, parent, risk):
        """:returns: An XML node with the details of an :obj:`euphorie.content.risk`."""
        node = etree.SubElement(parent, "risk", type=risk.type)
        if getattr(risk, "external_id", None):
            node.attrib["external-id"] = risk.external_id
        etree.SubElement(node, "title").text = risk.title
        etree.SubElement(node, "problem-description").text = StripUnwanted(
            risk.problem_description
        )
        node = self._add_string_or_html(node, risk.description, "description")
        if self.include_risk_legal_texts and StripMarkup(risk.legal_reference):
            node = self._add_string_or_html(
                node, risk.legal_reference, "legal-reference"
            )
        if self.is_etranslate_compatible:
            etree.SubElement(
                node,
                "show-not-applicable",
                attrib={"value": "true" if risk.show_notapplicable else "false"},
            )
        else:
            etree.SubElement(node, "show-not-applicable").text = (
                "true" if risk.show_notapplicable else "false"
            )
        if risk.type == "risk":
            method = etree.SubElement(node, "evaluation-method")
            method.text = risk.evaluation_method
            if risk.evaluation_method == "calculated":
                if risk.evaluation_algorithm() == "kinney":
                    if risk.default_probability:
                        method.attrib["default-probability"] = getToken(
                            IKinneyEvaluation["default_probability"],
                            risk.default_probability,
                            "0",
                        )
                    if risk.default_frequency:
                        method.attrib["default-frequency"] = getToken(
                            IKinneyEvaluation["default_frequency"],
                            risk.default_frequency,
                            "0",
                        )
                    if risk.default_effect:
                        method.attrib["default-effect"] = getToken(
                            IKinneyEvaluation["default_effect"],
                            risk.default_effect,
                            "0",
                        )
            elif risk.evaluation_method == "direct":
                if risk.default_priority:
                    method.attrib["default-priority"] = getToken(
                        IRisk["default_priority"], risk.default_priority
                    )

        for index in range(4):
            postfix = "" if not index else str(index + 1)
            image = getattr(risk, "image" + postfix, None)
            if image is not None:
                self.exportImage(node, image, getattr(risk, "caption" + postfix, None))

        if self.include_measures:
            solutions = [
                child for child in risk.values() if ISolution.providedBy(child)
            ]
            if solutions:
                sols = etree.SubElement(node, "solutions")
                for solution in solutions:
                    self.exportSolution(sols, solution)
        return node

    def exportSolution(self, parent, solution):
        """:returns: An XML node with the details
        of an :obj:`euphorie.content.solution`."""
        node = etree.SubElement(parent, "solution")
        if getattr(solution, "external_id", None):
            node.attrib["external-id"] = solution.external_id
        etree.SubElement(node, "description").text = StripUnwanted(solution.description)
        stripped_action = StripUnwanted(solution.action)
        if ISolution.providedBy(solution) and self.is_etranslate_compatible:
            solution_view = api.content.get_view(
                context=solution, name="nuplone-view", request=self.request
            )
            action_with_br = stripped_action.replace("\n", "<br/>")
            action_html = solution_view.render_md(action_with_br)
            fragment = html.fragment_fromstring(action_html, "action")
            node.append(fragment)
        else:
            etree.SubElement(node, "action").text = stripped_action
        if solution.requirements:
            etree.SubElement(node, "requirements").text = StripUnwanted(
                solution.requirements
            )
        return node

    def exportTrainingQuestion(self, parent, training_question):
        """:returns: An XML node with the details
        of an :obj:`euphorie.training_question`."""
        node = etree.SubElement(parent, "training_question")
        etree.SubElement(node, "title").text = StripUnwanted(training_question.title)
        etree.SubElement(node, "right_answer").text = StripUnwanted(
            training_question.right_answer
        )
        etree.SubElement(node, "wrong_answer_1").text = StripUnwanted(
            training_question.wrong_answer_1
        )
        etree.SubElement(node, "wrong_answer_2").text = StripUnwanted(
            training_question.wrong_answer_2
        )
        return node



from ..browser.survey import SurveyView
from euphorie.content.module import IModule
from euphorie.content.profilequestion import IProfileQuestion
from euphorie.content.survey import handleSurveyUnpublish
from euphorie.content.survey import Survey
from euphorie.testing import EuphorieIntegrationTestCase
from unittest import mock
from zope.annotation.interfaces import IAttributeAnnotatable
from zope.interface import alsoProvides
from zope.publisher.browser import TestRequest

import Acquisition
import unittest


class Mock(Acquisition.Explicit):
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def absolute_url(self):
        return "http://nohost/%s" % self.id

    def manage_fixupOwnershipAfterAdd(self):
        pass


class ViewTests(EuphorieIntegrationTestCase):
    def _request(self):
        req = TestRequest()
        alsoProvides(req, IAttributeAnnotatable)
        return req

    def test_update_no_children(self):
        survey = Survey()
        view = SurveyView(survey, self._request())
        self.assertEqual(view.modules_and_profile_questions, [])

    def test_update_with_profile(self):
        survey = Survey()
        child = Mock(id="child", title="Child")
        alsoProvides(child, IProfileQuestion)
        survey._setObject("child", child, suppress_events=True)
        view = SurveyView(survey, self._request())
        view._morph = mock.Mock(return_value="info")
        self.assertEqual(view.modules_and_profile_questions, ["info"])

    def test_update_with_module(self):
        survey = Survey()
        child = Mock(id="child", title="Child")
        alsoProvides(child, IModule)
        survey._setObject("child", child, suppress_events=True)
        view = SurveyView(survey, self._request())
        view._morph = mock.Mock(return_value="info")
        self.assertEqual(view.modules_and_profile_questions, ["info"])

    def test_update_other_child(self):
        survey = Survey()
        view = SurveyView(survey, self._request())
        child = Mock(id="child", title="Child")
        survey._setObject("child", child, suppress_events=True)
        self.assertEqual(view.modules_and_profile_questions, [])

    def test_moprh(self):
        child = Mock(id="child", title="Child")
        view = SurveyView(None, self._request())
        self.assertEqual(
            view._morph(child),
            {
                "id": "child",
                "title": "Child",
                "url": "http://nohost/child",
                "is_profile_question": False,
            },
        )


class HandleSurveyUnpublishTests(unittest.TestCase):
    def handleSurveyUnpublish(self, *a, **kw):
        return handleSurveyUnpublish(*a, **kw)

    def testRemovePublishedFromSurvey(self):
        surveygroup = Mock(published=None)
        surveygroup.survey = Mock(id="survey", published="yes")
        self.handleSurveyUnpublish(surveygroup.survey, None)
        self.assertTrue(not hasattr(surveygroup.survey, "published"))

    def testUpdatateSurveygroupIfCurrentlyPublished(self):
        surveygroup = Mock(published="survey")
        surveygroup.survey = Mock(id="survey", published="yes")
        self.handleSurveyUnpublish(surveygroup.survey, None)
        self.assertEqual(surveygroup.published, None)

    def testUpdatateSurveygroupIfOtherPublished(self):
        surveygroup = Mock(published="other")
        surveygroup.survey = Mock(id="survey", published="yes")
        self.handleSurveyUnpublish(surveygroup.survey, None)
        self.assertEqual(surveygroup.published, "other")



# htmllaundry was originally a library developed by Wichert Akkerman
#
# See: https://github.com/syslabcom/htmllaundry/tree/master
#
# A subset of the now unmaintained package code was copied and pasted
# into euphorie.htmllaundry
#
# The original htmllaundry package was licensed under the BSD license
#
